"""
Implementaion of the service parameter primitives.
"""
import codecs
import logging

from pydicom.uid import UID

from pynetdicom.pdu_items import (
    MaximumLengthSubItem,
    ImplementationClassUIDSubItem,
    ImplementationVersionNameSubItem,
    AsynchronousOperationsWindowSubItem,
    SCP_SCU_RoleSelectionSubItem,
    SOPClassExtendedNegotiationSubItem,
    SOPClassCommonExtendedNegotiationSubItem,
    UserIdentitySubItemRQ,
    UserIdentitySubItemAC
)
from pynetdicom.presentation import PresentationContext
from pynetdicom.utils import validate_ae_title
from pynetdicom._globals import DEFAULT_MAX_LENGTH

LOGGER = logging.getLogger('pynetdicom.pdu_primitives')


# TODO: Rename to UserInformation
class ServiceParameter(object):
    """ Base class for Service Parameters """

    def __eq__(self, other):
        """Equality of two ServiceParameters"""
        if isinstance(other, self.__class__):
            return other.__dict__ == self.__dict__

        return False

    def __ne__(self, other):
        """Inequality of two ServiceParameters"""
        return not self == other


# Association Service primitives
class A_ASSOCIATE(object):
    """
    An A-ASSOCIATE primitive.

    The establishment of an association between two AEs shall be performed
    through ACSE A-ASSOCIATE request, indication, response and confirmation
    primitives.

    The initiator of the service is called the Requestor and the user that
    receives the request is the Acceptor.

    The A-ASSOCIATE primitive is used by the DUL provider to send/receive
    information about the association. It gets converted to
    A-ASSOCIATE-RQ, -AC, -RJ PDUs that are sent to the peer DUL provider and
    gets deconverted from -RQ, -AC, -RJ PDUs received from the peer.

    It may be better to simply extend this with methods for containing
    the -rq, -ac, -rj possibilities rather than creating a new
    AssociationInformation class, but it would require maintaining the instance
    across the request-accept/reject path

    * -rq = no Result value
    * -ac = Result of 0x00
    * -rj = Result != 0x00

    +------------------+---------+------------+----------+--------------+
    | Parameter        | Request | Indication | Response | Confirmation |
    +------------------+---------+------------+----------+--------------+
    | app context name | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | calling ae title | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | called ae title  | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | user info        | M       | M(=)       | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | result           |         |            | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | source           |         |            |          | M            |
    +------------------+---------+------------+----------+--------------+
    | diagnostic       |         |            | U        | C(=)         |
    +------------------+---------+------------+----------+--------------+
    | calling pres add | M       | M(=)       |          |              |
    +------------------+---------+------------+----------+--------------+
    | called pres add  | M       | M(=)       |          |              |
    +------------------+---------+------------+----------+--------------+
    | pres contxt list | M       | M(=)       |          |              |
    +------------------+---------+------------+----------+--------------+
    | pres list result |         |            | M        | M(=)         |
    +------------------+---------+------------+----------+--------------+
    | mode             | UF      | MF(=)      |          |              |
    +------------------+---------+------------+----------+--------------+
    | resp ae title    |         |            | MF       | MF(=)        |
    +------------------+---------+------------+----------+--------------+
    | resp pres add    |         |            | MF       | MF(=)        |
    +------------------+---------+------------+----------+--------------+
    | pres/sess req    | UF      | UF(=)      | UF       | UF(=)        |
    +------------------+---------+------------+----------+--------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response


    The Requestor sends a request primitive to the local DICOM UL provider =>
    peer UL => indication primitive to Acceptor.

    Acceptor sends response primitive to peer UL => local UL => confirmation
    primitive to Requestor

    The DICOM UL providers communicate with UL users using service primitives
    The DICOM UL providers communicate with each other using PDUs over TCP/IP

    **Service Procedure**

    1. An AE (DICOM UL service user) that desires the establish an association
       issues an A-ASSOCIATE request primitive to the DICOM UL service
       provider. The Requestor shall not issue any primitives except the
       A-ABORT request primitive until it receives an A-ASSOCIATE confirmation
       primitive.
    2. The DICOM UL service provider issues an A-ASSOCIATE indication primitive
       to the called AE
    3. The called AE shall accept or reject the association by sending an
       A-ASSOCIATE response primitive with an appropriate Result parameter. The
       DICOM UL service provider shall issue an A-ASSOCIATE confirmation
       primitive having the same Result parameter. The Result Source parameter
       shall be assigned "UL service-user"
    4. If the Acceptor accepts the association, it is established and is
       available for use. DIMSE messages can now be exchanged.
    5. If the Acceptor rejects the association, it shall not be established and
       is not available for use
    6. If the DICOM UL service provider is not capable of supporting the
       requested association it shall return an A-ASSOCIATE confirmation
       primitive to the Requestor with an appropriate Result parameter
       (rejected). The Result Source parameter shall be assigned either
       UL service provider (ACSE) or UL service provider (Presentation).
       The indication primitive shall not be issued. The association shall not
       be established.
    7. Either Requestor or Acceptor may disrupt the Service Procedure by
       issuing an A-ABORT request primitive. The remote AE receives an A-ABORT
       indication primitive. The association shall not be established

    Attributes
    ----------
    mode : str
        Fixed value of "normal"
    application_context_name : pydicom.uid.UID, bytes or str
        The application context name proposed by the requestor. Acceptor
        returns either the same or a different name. Returned name specifies
        the application context used for the Association. See PS3.8 Annex A.
        The application context name shall be a valid UID or UID string and for
        version 3 of the DICOM Standard should be '1.2.840.10008.3.1.1.1'
    calling_ae_title : str or bytes
        Identifies the Requestor of the A-ASSOCIATE service. Must be a valid
        AE
    called_ae_title : str or bytes
        Identifies the intended Acceptor of the A-ASSOCIATE service. Must be a
        valid AE
    responding_ae_title : str or bytes
        Identifies the AE that contains the actual acceptor of the
        A-ASSOCIATE service. Shall always contain the same value as the
        Called AE Title of the A-ASSOCIATE indication
    user_information : list
        Used by Requestor and Acceptor to include AE user information. See
        PS3.8 Annex D and PS3.7 Annex D.3
    result : int
        Provided either by the Acceptor of the A-ASSOCIATE request, the UL
        service provider (ACSE related) or the UL service provider
        (Presentation related). Indicates the result of the A-ASSOCIATE
        service. Allowed values are:

        * 0: accepted
        * 1: rejected (permanent)
        * 2: rejected (transient)

    result_source : int
        Identifies the creating source of the Result and Diagnostic parameters
        Allowed values are:

        * 0: UL service-user
        * 1: UL service-provider (ACSE related function)
        * 2: UL service-provider (presentation related function)

    diagnostic : int
        If the `result` parameter is 0 "rejected (permanent)" or 1 "rejected
        (transient)" then this supplies diagnostic information about the
        result. If `result_source` is 0 "UL service-user" then allowed values
        are:

        * 0: no reason given
        * 1: application context name not supported
        * 2: calling AE title not recognised
        * 3: called AE title not recognised

        If `result_source` is 1 "UL service-provider (ACSE related function)"
        then allowed values are:

        * 0: no reason given
        * 1: no common UL version

        If `result_source` is 2 "UL service-provider (presentation related
        function)" then allowed values are:

        * 0: no reason given
        * 1: temporary congestion
        * 2: local limit exceeded
        * 3: called presentation address unknown
        * 4: presentation protocol version not supported
        * 5: no presentation service access point available

    calling_presentation_address : str
        TCP/IP address of the Requestor
    called_presentation_address : str
        TCP/IP address of the intended Acceptor
    responding_presentation_address : str
        Shall always contain the same value as the Called Presentation Address
    presentation_context_definition_list : list
        List of one or more presentation contexts, with each item containing
        a presentation context ID, an Abstract Syntax and a list of one or
        more Transfer Syntax Names. Sent by the Requestor during
        request/indication
    presentation_context_definition_results_list : list
        Used in response/confirmation to indicate acceptance or rejection of
        each presentation context definition.
        List of result values, with a one-to-one correspondence between each
        of the presentation contexts proposed in the Presentation Context
        Definition List parameter.
        The result values may be sent in any order and may be different than
        the order proposed.
        Only one Transfer Syntax per presentation context shall be agreed to
    presentation_requirements : str
        Fixed value of "Presentation Kernel"
    session_requirements : str
        Fixed value of "" (empty string)

    References
    ----------

    * DICOM Standard, Part 8, Section 7.1.1
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        self.application_context_name = None
        self.calling_ae_title = None
        self.called_ae_title = None
        self.user_information = []
        self.result = None
        self.result_source = None
        self.diagnostic = None
        self.calling_presentation_address = None
        self.called_presentation_address = None
        self.presentation_context_definition_list = []
        self.presentation_context_definition_results_list = []

    @property
    def mode(self):
        """Return the Mode parameter."""
        return "normal"

    @property
    def application_context_name(self):
        """Return the Application Context Name parameter."""
        return self._application_context_name

    @application_context_name.setter
    def application_context_name(self, value):
        """Set the Application Context Name parameter.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The value for the Application Context Name
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, UID):
            pass
        elif isinstance(value, str):
            value = UID(value)
        elif isinstance(value, bytes):
            value = UID(value.decode('ascii'))
        elif value is None:
            pass
        else:
            raise TypeError("application_context_name must be a "
                            "pydicom.uid.UID, str or bytes")

        if value is not None and not value.is_valid:
            LOGGER.error("application_context_name is an invalid UID")
            raise ValueError("application_context_name is an invalid UID")

        self._application_context_name = value

    @property
    def calling_ae_title(self):
        """Return the Calling AE Title parameter."""
        return self._calling_ae_title

    @calling_ae_title.setter
    def calling_ae_title(self, value):
        """Set the Calling AE Title parameter.

        Parameters
        ----------
        value : str or bytes
            The Calling AE Title as a string or bytes object. Cannot be an
            empty string and will be truncated to 16 characters long
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, str):
            value = codecs.encode(value, 'ascii')

        if value is not None:
            self._calling_ae_title = validate_ae_title(value)
        else:
            self._calling_ae_title = None

    @property
    def called_ae_title(self):
        """Return the Called AE Title parameter."""
        return self._called_ae_title

    @called_ae_title.setter
    def called_ae_title(self, value):
        """Set the Called AE Title parameter.

        Parameters
        ----------
        value : str or bytes
            The Called AE Title as a string or bytes object. Cannot be an empty
            string and will be truncated to 16 characters long
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, str):
            value = codecs.encode(value, 'ascii')

        if value is not None:
            self._called_ae_title = validate_ae_title(value)
        else:
            self._called_ae_title = None

    @property
    def responding_ae_title(self):
        """Return the Responding AE Title parameter."""
        return self.called_ae_title

    @property
    def user_information(self):
        """Return the User Information parameter."""
        return self._user_information

    @user_information.setter
    def user_information(self, value_list):
        """Set the A-ASSOCIATE primitive's User Information parameter.

        Parameters
        ----------
        value_list : list of user information class objects
            A list of user information objects, must contain at least
            MaximumLengthNotification and ImplementationClassUIDNotification
        """
        # pylint: disable=attribute-defined-outside-init
        valid_usr_info_items = []

        if isinstance(value_list, list):
            # Iterate through the items and check they're an acceptable class
            for item in value_list:
                if item.__class__.__name__ in \
                        ["MaximumLengthNotification",
                         "ImplementationClassUIDNotification",
                         "ImplementationVersionNameNotification",
                         "AsynchronousOperationsWindowNegotiation",
                         "SCP_SCU_RoleSelectionNegotiation",
                         "SOPClassExtendedNegotiation",
                         "SOPClassCommonExtendedNegotiation",
                         "UserIdentityNegotiation"]:
                    valid_usr_info_items.append(item)
                else:
                    LOGGER.info("Attempted to set "
                                "A_ASSOCIATE.user_information to a list "
                                "which includes an unsupported item")
        else:
            LOGGER.error("A_ASSOCIATE.user_information must be a list")
            raise TypeError("A_ASSOCIATE.user_information must be a list")

        self._user_information = valid_usr_info_items

    @property
    def result(self):
        """Return te Result parameter."""
        return self._result

    @result.setter
    def result(self, value):
        """Set the A-ASSOCIATE Service primitive's Result parameter.

        Parameters
        ----------
        value : str
            One of the following:

            * 0: accepted
            * 1: rejected (permanent)
            * 2: rejected (transient)
        """
        # pylint: disable=attribute-defined-outside-init
        if value is None:
            pass
        elif value not in [0, 1, 2]:
            LOGGER.error("A_ASSOCIATE.result set to an unknown value")
            raise ValueError("Unknown A_ASSOCIATE.result value")

        self._result = value

    @property
    def result_source(self):
        """Return the Result Source parameter."""
        return self._result_source

    @result_source.setter
    def result_source(self, value):
        """Set the A-ASSOCIATE Service primitive's Result Source parameter.

        Parameters
        ----------
        value : int
            One of the following:

            * 1: UL service-user
            * 2: UL service-provider (ACSE related function)
            * 3: UL service-provider (presentation related function)
        """
        # pylint: disable=attribute-defined-outside-init
        if value is None:
            pass
        elif value not in [1, 2, 3]:
            LOGGER.error("A_ASSOCIATE.result_source set to an unknown value")
            raise ValueError("Unknown A_ASSOCIATE.result_source value")

        self._result_source = value

    @property
    def diagnostic(self):
        """Return the Diagnostic parameter."""
        return self._diagnostic

    @diagnostic.setter
    def diagnostic(self, value):
        """
        Set the A-ASSOCIATE Service primitive's Diagnostic parameter

        Parameters
        ----------
        value : int
            If `result_source` is "UL service-user" then allowed values are:

            * 1: no reason given
            * 2: application context name not supported
            * 3: calling AE title not recognised
            * 7: called AE title not recognised

            If `result_source` is "UL service-provider (ACSE related function)"
            then allowed values are:

            * 1: no reason given
            * 2: protocol version not supported"

            If `result_source` is "UL service-provider (Presentation related
            function)" then allowed values are:

            * 1: temporary congestion
            * 2: local limit exceeded
        """
        # pylint: disable=attribute-defined-outside-init
        if value is None:
            pass
        elif value not in [1, 2, 3, 7]:
            LOGGER.error("A_ASSOCIATE.diagnostic set to an unknown value")
            raise ValueError("Unknown A_ASSOCIATE.diagnostic value")

        self._diagnostic = value

    @property
    def calling_presentation_address(self):
        """Return the Calling Presentation Address parameter."""
        return self._calling_presentation_address

    @calling_presentation_address.setter
    def calling_presentation_address(self, value):
        """
        Set the A-ASSOCIATE Service primitive's Calling Presentation
        Address parameter

        Parameters
        ----------
        value : (str, int) tuple
            A tuple containing a valid TCP/IP address string and the port
            number as an int
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, tuple):
            if len(value) == 2 and isinstance(value[0], str) \
                    and isinstance(value[1], int):
                self._calling_presentation_address = value
            else:
                LOGGER.error("A_ASSOCIATE.calling_presentation_address must "
                             "be (str, int) tuple")
                raise TypeError("A_ASSOCIATE.calling_presentation_address "
                                "must be (str, int) tuple")
        elif value is None:
            self._calling_presentation_address = value
        else:
            LOGGER.error("A_ASSOCIATE.calling_presentation_address must be "
                         "(str, int) tuple")
            raise TypeError("A_ASSOCIATE.calling_presentation_address must "
                            "be (str, int) tuple")

    @property
    def called_presentation_address(self):
        """Return the Called Presentation Address parameter."""
        return self._called_presentation_address

    @called_presentation_address.setter
    def called_presentation_address(self, value):
        """Set the Called Presentation Address parameter.

        Parameters
        ----------
        value : (str, int) tuple
            A tuple containing a valid TCP/IP address string and the port
            number as an int
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, tuple):
            if len(value) == 2 and isinstance(value[0], str) \
                    and isinstance(value[1], int):
                self._called_presentation_address = value
            else:
                LOGGER.error("A_ASSOCIATE.called_presentation_address must "
                             "be (str, int) tuple")
                raise TypeError("A_ASSOCIATE.called_presentation_address "
                                "must be (str, int) tuple")
        elif value is None:
            self._called_presentation_address = value
        else:
            LOGGER.error("A_ASSOCIATE.called_presentation_address must be "
                         "(str, int) tuple")
            raise TypeError("A_ASSOCIATE.called_presentation_address must "
                            "be (str, int) tuple")

    @property
    def responding_presentation_address(self):
        """Get the Responding Presentation Address parameter."""
        return self.called_presentation_address

    @property
    def presentation_context_definition_list(self):
        """Get the Presentation Context Definition List."""
        return self._presentation_context_definition_list

    @presentation_context_definition_list.setter
    def presentation_context_definition_list(self, value_list):
        """
        Set the A-ASSOCIATE Service primitive's Presentation Context Definition
        List parameter

        Parameters
        ----------
        value_list : list of utils.PresentationContext
            The Presentation Contexts proposed by the Association Requestor
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value_list, list):
            valid_items = []
            for item in value_list:
                if isinstance(item, PresentationContext):
                    valid_items.append(item)
                else:
                    LOGGER.warning(
                        "Attempted to set "
                        "A_ASSOCIATE.presentation_context_definition_list to "
                        "a list which includes an invalid items"
                    )

            self._presentation_context_definition_list = valid_items

        else:
            LOGGER.error("A_ASSOCIATE.presentation_context_definition_list "
                         "must be a list")
            raise TypeError("A_ASSOCIATE.presentation_context_definition_list "
                            "must be a list")

    @property
    def presentation_context_definition_results_list(self):
        """Get the Presentation Context Definition Results List."""
        return self._presentation_context_definition_results_list

    @presentation_context_definition_results_list.setter
    def presentation_context_definition_results_list(self, value_list):
        """Set the Presentation Context Definition Results List parameter.

        Parameters
        ----------
        value_list : list of utils.PresentationContext
            The results of the Presentation Contexts proposal by the
            Association Requestor
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value_list, list):
            valid_items = []
            for item in value_list:
                if isinstance(item, PresentationContext):
                    valid_items.append(item)
                else:
                    LOGGER.warning("Attempted to set A_ASSOCIATE.presentation"
                                   "_context_definition_results_list to a "
                                   "list which includes one or more invalid "
                                   "items.")

            self._presentation_context_definition_results_list = valid_items

        else:
            LOGGER.error("A_ASSOCIATE.presentation_context_definition_"
                         "results_list must be a list")
            raise TypeError("A_ASSOCIATE.presentation_context_definition_"
                            "results_list must be a list")

    @property
    def presentation_requirements(self):
        """Get the Presentation Kernel."""
        return "Presentation Kernel"

    @property
    def session_requirements(self):
        """Get the Session Requirements."""
        return ""

    # Shortcut attributes for User Information items
    # Mandatory UI Items
    @property
    def maximum_length_received(self):
        """Get the Maximum Length Received."""
        for item in self.user_information:
            if isinstance(item, MaximumLengthNotification):
                return item.maximum_length_received

        return None

    @maximum_length_received.setter
    def maximum_length_received(self, value):
        """Set the Maximum Length Received.

        If the A_ASSOCIATE.user_information list contains a
        MaximumLengthNotification item then set its maximum_length_received
        value. If not then add a MaximumLengthNotification item and set its
        maximum_length_received value.

        Parameters
        ----------
        value : int
            The maximum length of each P-DATA in bytes
        """
        # Type and value checking for the maximum_length_received parameter is
        #   done by the MaximumLengthNotification class

        # Check for a MaximumLengthNotification item
        found_item = False

        for item in self.user_information:
            if isinstance(item, MaximumLengthNotification):
                found_item = True
                item.maximum_length_received = value

        # No MaximumLengthNotification item found
        if not found_item:
            max_length = MaximumLengthNotification()
            max_length.maximum_length_received = value
            self.user_information.append(max_length)

    @property
    def implementation_class_uid(self):
        """Return the Implementation Class UID."""
        for item in self.user_information:
            if isinstance(item, ImplementationClassUIDNotification):
                if item.implementation_class_uid is None:
                    LOGGER.error("Implementation Class UID has not been set")
                    raise ValueError("Implementation Class UID has not "
                                     "been set")

                return item.implementation_class_uid

        LOGGER.error("Implementation Class UID has not been set")
        raise ValueError("Implementation Class UID has not been set")

    @implementation_class_uid.setter
    def implementation_class_uid(self, value):
        """Set the Implementation Class UID.

        If the A_ASSOCIATE.user_information list contains an
        ImplementationClassUIDNotification item then set its
        implementation_class_uid value. If not then add a
        ImplementationClassUIDNotification item and set its
        implementation_class_uid value.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The value for the Implementation Class UID
        """
        # Type and value checking for the implementation_class_uid parameter is
        #   done by the ImplementationClassUIDNotification class

        # Check for a ImplementationClassUIDNotification item
        found_item = False
        for item in self.user_information:
            if isinstance(item, ImplementationClassUIDNotification):
                found_item = True
                item.implementation_class_uid = value

        # No ImplementationClassUIDNotification item found
        if not found_item:
            imp_uid = ImplementationClassUIDNotification()
            imp_uid.implementation_class_uid = value
            self.user_information.append(imp_uid)


class A_RELEASE(object):
    """
    An A-RELEASE primitive.

    The release of an association between two AEs shall be performed through
    ACSE A-RELEASE request, indication, response and confirmation primitives.
    The initiator of the service is called a Requestor and the service-user
    that receives the A-RELEASE indication is called the acceptor.

    Service Procedure

    1. The user (Requestor) that desires to end the association issues an
       A-RELEASE request primitive. The Requestor shall not issue any other
       primitives other than A-ABORT until it receives an A-RELEASE
       confirmation primitive.
    2. The DUL provider issues an A-RELEASE indication to the Acceptor. The
       Acceptor shall not issue any other primitives other than A-RELEASE
       response, A-ABORT request or P-DATA request.
    3. To complete the release, the Acceptor replies using an A-RELEASE
       response primitive, with "affirmative" as the result parameter.
    4. After the Acceptor issues the A-RELEASE response it shall not issue any
       more primitives.
    5. The Requestor shall issue an A-RELEASE confirmation primitive always
       with an "affirmative" value for the Result parameter.
    6. A user may disrupt the release by issuing an A-ABORT request.
    7. A collision may occur when both users issue A-RELEASE requests
       simultaneously. In this situation both users receive an unexpect
       A-RELEASE indication primitive (instead of an A-RELEASE acceptance):

         a. The association requestor issues an A-RELEASE response primitive
         b. The association acceptor waits for an A-RELEASE confirmation
            primitive from its peer. When it receives one it issues an
            A-RELEASE response primitive
         c. The association requestor receives an A-RELEASE confirmation
            primitive.

       When both ACSE users have received an A-RELEASE confirmation primitive
       the association shall be released.

    +------------------+---------+------------+----------+--------------+
    | Parameter        | Request | Indication | Response | Confirmation |
    +------------------+---------+------------+----------+--------------+
    | reason           | UF      | UF(=)      | UF       | UF(=)        |
    +------------------+---------+------------+----------+--------------+
    | user info        |NU       | NU(=)      | NU       | NU(=)        |
    +------------------+---------+------------+----------+--------------+
    | result           |         |            | MF       | MF(=)        |
    +------------------+---------+------------+----------+--------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response

    Attributes
    ----------
    reason : str
        Fixed value of "normal". Identifies the general level of urgency of the
        request.
    result : str or None
        Must be None for request and indication, "affirmative" for response
        and confirmation.

    References
    ----------
    * DICOM Standard, Part 8, Section 7.2
    """
    def __init__(self):
        self.result = None

    @property
    def reason(self):
        """Return the *Reason* parameter."""
        return "normal"

    @property
    def result(self):
        """Return the *Result* parameter."""
        return self._result

    @result.setter
    def result(self, value):
        """Set the Result parameter."""
        # pylint: disable=attribute-defined-outside-init
        if value is not None and value != "affirmative":
            LOGGER.error("A_RELEASE.result must be None or 'affirmative'")
            raise ValueError("A_RELEASE.result must be None or 'affirmative'")

        self._result = value


class A_ABORT(object):
    """
    An A-ABORT primitive.

    +------------------+---------+------------+
    | Parameter        | Request | Indication |
    +------------------+---------+------------+
    | abort source     |         | M          |
    +------------------+---------+------------+
    | user info        |NU       | NU(=)      |
    +------------------+---------+------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response

    Attributes
    ----------
    abort_source : int
        Indicates the initiating source of the abort. Allowed values are:

        * 0: UL service-user
        * 2: UL service-provider

    References
    ----------

    * DICOM Standard, Part 8, Section 7.3.1
    """

    def __init__(self):
        self.abort_source = None

    @property
    def abort_source(self):
        """Return the *Abort Source*."""
        if self._abort_source is None:
            LOGGER.error("A_ABORT.abort_source parameter not set")
            raise ValueError("A_ABORT.abort_source value not set")

        return self._abort_source

    @abort_source.setter
    def abort_source(self, value):
        """Set the Abort Source."""
        # pylint: disable=attribute-defined-outside-init
        if value in [0, 2]:
            self._abort_source = value
        elif value is None:
            self._abort_source = None
        else:
            LOGGER.error("Attempted to set A_ABORT.abort_source to an "
                         "invalid value")
            raise ValueError("Attempted to set A_ABORT.abort_source to an "
                             "invalid value")


class A_P_ABORT(object):
    """
    An A-P-ABORT primitive.

    +------------------+------------+
    | Parameter        | Indication |
    +------------------+------------+
    | abort source     | P          |
    +------------------+------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | P   - Provider initiated
    | (=) - shall have same value as request or response

    Attributes
    ----------
    provider_reason : int
        Indicates the reason for the abort. Allowed values are:

        * 0: reason not specified
        * 1: unrecognised PDU
        * 2: unexpected PDU
        * 4: unrecognised PDU parameter
        * 5: unexpected PDU parameter
        * 6: invalid PDU parameter value

    References
    ----------

    * DICOM Standard, Part 8, Section 7.4.1
    """

    def __init__(self):
        self.provider_reason = None

    @property
    def provider_reason(self):
        """Return the *Provider Reason*."""
        if self._provider_reason is None:
            LOGGER.error("A_ABORT.provider_reason parameter not set")
            raise ValueError("A_ABORT.provider_reason value not set")

        return self._provider_reason

    @provider_reason.setter
    def provider_reason(self, value):
        """Set the Provider Reason."""
        # pylint: disable=attribute-defined-outside-init
        if value in [0, 1, 2, 4, 5, 6]:
            self._provider_reason = value
        elif value is None:
            self._provider_reason = None
        else:
            msg = (
                "Attempted to set A_P_ABORT.provider_reason to an invalid "
                "value"
            )
            LOGGER.error(msg)
            raise ValueError(msg)


class P_DATA(object):
    """
    A P-DATA primitive.

    +------------------------------+---------+------------+
    | Parameter                    | Request | Indication |
    +------------------------------+---------+------------+
    | presentation data value list | M       | M(=)       |
    +------------------------------+---------+------------+

    | U   - User option
    | UF  - User option, fixed value
    | C   - Conditional (on user option)
    | M   - Mandatory
    | MF  - Mandatory, fixed value
    | NU  - Not used
    | (=) - shall have same value as request or response

    Attributes
    ----------
    presentation_data_value_list : list of [int, bytes]
        Contains one or more Presentation Data Values (PDV), each consisting of
        a Presentation Context ID and User Data values. The User Data values
        are taken from the Abstract Syntax and encoded in the Transfer Syntax
        identified by the Presentation Context ID. Each item in the list is
        [Context ID, PDV Data]

    References
    ----------

    * DICOM Standard, Part 8, Section 7.6.1
    """
    def __init__(self):
        self.presentation_data_value_list = []

    @property
    def presentation_data_value_list(self):
        """Return the *Presentation Data Value List*."""
        return self._presentation_data_value_list

    @presentation_data_value_list.setter
    def presentation_data_value_list(self, value_list):
        """Set the Presentation Data Value List."""
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value_list, list):
            for pdv in value_list:
                if isinstance(pdv, list):
                    if isinstance(pdv[0], int) and isinstance(pdv[1], bytes):
                        pass
                    else:
                        raise TypeError("P_DATA.presentation_data_value_list "
                                        "should be a list of [int, bytes]")
                else:
                    raise TypeError("P_DATA.presentation_data_value_list "
                                    "should be a list of [ID, PDV]")
        else:
            raise TypeError("P_DATA.presentation_data_value_list "
                            "should be a list of [int, bytes]")

        self._presentation_data_value_list = value_list

    def __str__(self):
        """String representation of the class."""
        s = 'P-DATA\n'
        for pdv in self.presentation_data_value_list:
            s += '  Context ID: {0!s}\n'.format(pdv[0])
            s += '  Value Length: {0!s} bytes\n'.format(len(pdv[1]))
            header_byte = pdv[1][0]

            # Python 2 compatibility
            if isinstance(header_byte, str):
                header_byte = ord(header_byte)

            s += "  Message Control Header Byte: {:08b}\n".format(header_byte)

            # xxxxxx01 and xxxxxx011
            if header_byte & 1:
                # xxxxxx11
                if header_byte & 2:
                    s += '    Command information, last fragment of the ' \
                         'DIMSE message\n'
                # xxxxxx01
                else:
                    s += '    Command information, not the last fragment of ' \
                         'the DIMSE message\n'
            # xxxxxx00, xxxxxxx10
            else:
                # xxxxxx10
                if header_byte & 2 != 0:
                    s += '    Dataset information, last fragment of the ' \
                         'DIMSE message\n'
                # xxxxxx00
                else:
                    s += '    Dataset information, not the last fragment of ' \
                         'the DIMSE message\n'

            # Remaining data
            #s += pretty_bytes(pdv[1][1:], '    ', max_size=512)

        return s


# User Information Negotiation primitives
class MaximumLengthNotification(ServiceParameter):
    """
    A representation of a Maximum Length Negotiation primitive.

    The maximum length notification allows communicating AEs to limit the size
    of the data for each P-DATA indication. This notification is required for
    all DICOM v3.0 conforming implementations.

    This User Information item is required during Association negotiation and
    there must only be a single MaximumLengthNotification item

    Attributes
    ----------
    maximum_length_received : int
        The maximum length received value for the Maximum Length sub-item in
        bytes. A value of 0 indicates unlimited length (31682 bytes default).

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.1
    * DICOM Standard, Part 8, Annex D.1
    """
    def __init__(self):
        self.maximum_length_received = DEFAULT_MAX_LENGTH

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.MaximumLengthSubItem
        """
        item = MaximumLengthSubItem()
        item.from_primitive(self)

        return item

    @property
    def maximum_length_received(self):
        """Return the *Maximum Length Received*."""
        return self._maximum_length

    @maximum_length_received.setter
    def maximum_length_received(self, val):
        """User defined Maximum Length to be used during an Association.

        Parameters
        ----------
        val : int
            The maximum length of each P-DATA in bytes, must be equal to or
            greater than 0. A value of 0 indicates an unlimited maximum length.

        Raises
        ------
        ValueError
            If `maximum_length_received` is negative
        TypeError
            If `maximum_length_received` is not an int
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(val, int):
            if val < 0:
                LOGGER.error('Maximum Length Received must be greater than 0')
                raise ValueError("Maximum Length Received must be greater "
                                 "than 0")
            else:
                self._maximum_length = val
        else:
            LOGGER.error("Maximum Length Received must be numerical")
            raise TypeError("Maximum Length Received must be numerical")

    def __str__(self):
        """String representation of the class."""
        s = "Maximum Length Negotiation\n"
        s += "  Maximum length received: {0:d} bytes\n".format(
            self.maximum_length_received)
        return s


# TODO: Combine ImplementationClass and ImplementationVersion
#   into ImplementationIdentificationNotification
class ImplementationClassUIDNotification(ServiceParameter):
    """A representation of a Implementation Class UID Notification primitive.

    The implementation identification notification allows implementations of
    communicating AEs to identify each other at Association establishment time.
    It is intended to provider respective and non-ambiguous identification in
    the event of communication problems encountered between two nodes. This
    negotiation is required.

    Implementation identification relies on two pieces of information:
    - Implementation Class UID (required)
    - Implementation Version Name (optional)

    The Implementation Class UID is required during Association negotiation and
    there must only be a single ImplementationClassUID item

    usr_data_neg = []
    usr_data_neg.append(impl_class_uid)

    Attributes
    ----------
    implementation_class_uid : pydicom.uid.UID, bytes or str
        The UID to use

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import (
    ...     ImplementationClassUIDNotification
    ... )
    >>> item = ImplementationClassUIDNotification()
    >>> item.implementation_class_uid = '1.2.3.4'

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.2
    """
    def __init__(self):
        self.implementation_class_uid = None

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.ImplementationClassUIDSubItem

        Raises
        ------
        ValueError
            If no UID is set
        """
        if self.implementation_class_uid is None:
            LOGGER.error("The Implementation Class UID must be set prior to "
                         "requesting Association")
            raise ValueError("The Implementation Class UID must be set "
                             "prior to requesting Association")

        item = ImplementationClassUIDSubItem()
        item.from_primitive(self)

        return item

    @property
    def implementation_class_uid(self):
        """Return the *Implementation Class UID*."""
        return self._implementation_class_uid

    @implementation_class_uid.setter
    def implementation_class_uid(self, value):
        """Sets the Implementation Class UID parameter.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The value for the Implementation Class UID
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, UID):
            pass
        elif isinstance(value, str):
            value = UID(value)
        elif isinstance(value, bytes):
            value = UID(value.decode('ascii'))
        elif value is None:
            pass
        else:
            raise TypeError("Implementation Class UID must be a "
                            "pydicom.uid.UID, str or bytes")

        if value is not None and not value.is_valid:
            LOGGER.error(
                "The Implementation Class UID Notification's 'Implementation "
                "Class UID' parameter value '{}' is not a valid UID"
                .format(value)
            )

        self._implementation_class_uid = value

    def __str__(self):
        """String representation of the class."""
        s = "Implementation Class UID\n"
        s += "  Implementation class UID: {0!s}\n" \
             .format(self.implementation_class_uid)
        return s


class ImplementationVersionNameNotification(ServiceParameter):
    """
    A representation of a Implementation Version Name Notification primitive.

    The implementation identification notification allows implementations of
    communicating AEs to identify each other at Association establishment time.
    It is intended to provider respective and non-ambiguous identification in
    the event of communication problems encountered between two nodes. This
    negotiation is required.

    Implementation identification relies on two pieces of information:

    - Implementation Class UID (required)
    - Implementation Version Name (optional)

    The Implementation Version Name is optional and there may only be a single
    ImplementationVersionName item

    Attributes
    ----------
    implementation_version_name : str or bytes
        The version name to use, maximum of 16 characters

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import (
    ...     ImplementationVersionNameNotification
    ... )
    >>> item = ImplementationVersionNameNotification()
    >>> item.implementation_version_name = b'SOME_NAME'

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.2
    """
    def __init__(self):
        self.implementation_version_name = None

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.ImplementationVersionNameSubItem

        Raises
        ------
        ValueError
            If no name is set
        """
        if self.implementation_version_name is None:
            raise ValueError("Implementation Version Name must be set prior "
                             "to Association")

        item = ImplementationVersionNameSubItem()
        item.from_primitive(self)

        return item

    @property
    def implementation_version_name(self):
        """Return the *Implementation Version Name*."""
        return self._implementation_version_name

    @implementation_version_name.setter
    def implementation_version_name(self, value):
        """Sets the Implementation Version Name parameter.

        Parameters
        ----------
        value : str or bytes
            The value for the Implementation Version Name

        Raises
        ------
        TypeError
            If `value` is not a str or bytes
        ValueError
            If `value` is empty or longer than 16 characters
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, str):
            value = codecs.encode(value, 'ascii')
        elif isinstance(value, bytes):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("Implementation Version Name must be a str or bytes")
            raise TypeError("Implementation Version Name must be a str "
                            "or bytes")

        if value is not None and not 1 < len(value) < 17:
            raise ValueError("Implementation Version Name must be "
                             "between 1 and 16 characters long")

        self._implementation_version_name = value

    def __str__(self):
        """String representation of the class."""
        s = "Implementation Version Name\n"
        s += "  Implementation version name: {0!s}\n".format(
            self.implementation_version_name)
        return s


class AsynchronousOperationsWindowNegotiation(ServiceParameter):
    """
    Representation of the Asynchronous Operations Window Negotiation primitive.

    Allows peer AEs to negotiate the maximum number of outstanding operation
    or sub-operation requests. This negotiation is optional.

    The Asynchronous Operations Window is optional and there may only be a
    single AsynchronousOperationsWindowNegotiation item

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC

    Attributes
    ----------
    maximum_number_operations_invoked : int
        The maximum number of asynchronous operations invoked by the AE. A
        value of 0 indicates unlimited operations (default 1)
    maximum_number_operations_performed : int
        The maximum number of asynchronous operations performed by the AE. A
        value of 0 indicates unlimited operations (default 1)

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import (
    ...     AsynchronousOperationsWindowNegotiation
    ... )
    >>> item = AsynchronousOperationsWindowNegotiation()
    >>> item.maximum_number_operations_invoked = 2
    >>> item.maximum_number_operations_performed = 1

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.3
    """

    def __init__(self):
        self.maximum_number_operations_invoked = 1
        self.maximum_number_operations_performed = 1

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.AsynchronousOperationsWindowSubItem
        """
        item = AsynchronousOperationsWindowSubItem()
        item.from_primitive(self)

        return item

    @property
    def maximum_number_operations_invoked(self):
        """Return the *Maximum Number Operations Invoked*."""
        return self._maximum_number_operations_invoked

    @maximum_number_operations_invoked.setter
    def maximum_number_operations_invoked(self, value):
        """Sets the Maximum Number Operations Invoked parameter.

        Parameters
        ----------
        value : int
            The maximum number of operations invoked

        Raises
        ------
        TypeError
            If `value` is not an int
        ValueError
            If `value` is less than 0
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, int):
            pass
        else:
            LOGGER.error("Maximum Number Operations Invoked must be an int")
            raise TypeError("Maximum Number Operations Invoked must be an int")

        if value < 0:
            raise ValueError("Maximum Number Operations Invoked must be "
                             "greater than 0")

        self._maximum_number_operations_invoked = value

    @property
    def maximum_number_operations_performed(self):
        """Return the *Maximum Number Operations Performed*."""
        return self._maximum_number_operations_performed

    @maximum_number_operations_performed.setter
    def maximum_number_operations_performed(self, value):
        """
        Sets the Maximum Number Operations Performed parameter

        Parameters
        ----------
        value : int
            The maximum number of operations performed

        Raises
        ------
        TypeError
            If `value` is not an int
        ValueError
            If `value` is less than 0
        """
        # pylint: disable=attribute-defined-outside-init
        if not isinstance(value, int):
            LOGGER.error("Maximum Number Operations Performed must be an int")
            raise TypeError("Maximum Number Operations Performed must be "
                            "an int")

        if value < 0:
            raise ValueError("Maximum Number Operations Performed must be "
                             "greater than 0")

        self._maximum_number_operations_performed = value

    def __str__(self):
        """String representation of the class."""
        s = "Asynchronous Operations Window\n"
        s += "  Maximum number operations invoked: {0:d}\n".format(
            self.maximum_number_operations_invoked)
        s += "  Maximum number operations performed: {0:d}\n".format(
            self.maximum_number_operations_performed)
        return s


class SCP_SCU_RoleSelectionNegotiation(ServiceParameter):
    """
    A representation of the SCP/SCU Role Selection Negotiation primitive.

    Allows peer AEs to negotiate the roles in which they will serve for each
    SOP Class or Meta SOP Class supported on the Association. This negotiation
    is optional.

    The Association Requestor may use one SCP/SCU Role Selection item for each
    SOP Class as identified by its corresponding Abstract Syntax Name and shall
    be one of three role values:

    - Requestor is SCU only
    - Requestor is SCP only
    - Requestor is both SCU/SCP

    If the SCP/SCU Role Selection item is absent the default role for a
    Requestor is SCU and for an Acceptor is SCP.

    For a Requestor support for each SOP Class shall be one of the following
    roles:

    * Requestor is SCU only
    * Requestor is SCP only
    * Requestor is both SCU and SCP

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC

    Attributes
    ----------
    sop_class_uid : pydicom.uid.UID, bytes or str
        The UID of the corresponding Abstract Syntax
    scu_role : bool
        False for non-support of the SCU role, True for support
    scp_role : bool
        False for non-support of the SCP role, True for support

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation
    >>> item = SCP_SCU_RoleSelectionNegotiation()
    >>> item.sop_class_uid = '1.2.840.10008.5.1.4.1.1.2'
    >>> item.scu_role = True
    >>> item.scp_role = False

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.4
    """
    def __init__(self):
        self.sop_class_uid = None
        self.scu_role = None
        self.scp_role = None

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded

        Returns
        -------
        item : pdu_items.SCP_SCU_RoleSelectionSubItem

        Raises
        ------
        ValueError
            If no SOP Class UID, SCU Role or SCP Role is set
        ValueError
            If SCU Role and SCP Role are both False
        """
        if self.sop_class_uid is None:
            LOGGER.error("'sop_class_uid' must be set prior to Association")
            raise ValueError(
                "'sop_class_uid' must be set prior to Association"
            )

        # To get to this point self.sop_class_uid must be set
        if not self.scu_role and not self.scp_role:
            LOGGER.error("SCU and SCP Roles cannot both be unsupported "
                         "for %s", self.sop_class_uid)
            raise ValueError("SCU and SCP Roles cannot both be unsupported "
                             "for {}".format(self.sop_class_uid))

        item = SCP_SCU_RoleSelectionSubItem()
        item.from_primitive(self)

        return item

    @property
    def scp_role(self):
        """Return the *SCP Role*."""
        return self._scp_role

    @scp_role.setter
    def scp_role(self, value):
        """Sets the SCP Role parameter.

        Parameters
        ----------
        value : bool
            True if supported, False otherwise (default)

        Raises
        ------
        TypeError
            If `value` is not a bool
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bool):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("SCP Role must be boolean")
            raise TypeError("SCP Role must be boolean")

        self._scp_role = value

    @property
    def scu_role(self):
        """Return the *SCU Role*."""
        return self._scu_role

    @scu_role.setter
    def scu_role(self, value):
        """Sets the SCU Role parameter.

        Parameters
        ----------
        value : bool
            True if supported, False otherwise

        Raises
        ------
        TypeError
            If `value` is not a bool
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bool):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("SCU Role must be boolean")
            raise TypeError("SCU Role must be boolean")

        self._scu_role = value

    @property
    def sop_class_uid(self):
        """Return the *SOP Class UID*."""
        return self._sop_class_uid

    @sop_class_uid.setter
    def sop_class_uid(self, value):
        """Sets the SOP Class UID parameter.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The corresponding Abstract Syntax UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, UID):
            pass
        elif isinstance(value, str):
            value = UID(value)
        elif isinstance(value, bytes):
            value = UID(value.decode('ascii'))
        elif value is None:
            pass
        else:
            LOGGER.error("SOP Class UID must be a pydicom.uid.UID, str "
                         "or bytes")
            raise TypeError("SOP Class UID must be a pydicom.uid.UID, str "
                            "or bytes")

        if value is not None and not value.is_valid:
            LOGGER.error("Implementation Class UID is an invalid UID")
            raise ValueError("Implementation Class UID is an invalid UID")

        self._sop_class_uid = value


class SOPClassExtendedNegotiation(ServiceParameter):
    """
    A representation of the SOP Class Extended Negotiation primitive.

    Allows peer AEs to exchange application information defined by specific
    Service Class specifications. Each Service Class is required to document
    the application information it supports and how this information is
    negotiated between SCUs and SCPs.

    The SOP Class Extended Negotiation is optional and there may only be a
    single SOPClassExtendedNegotiation item for each available SOP Class UID.

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC

    Attributes
    ----------
    sop_class_uid : pydicom.uid.UID, bytes or str
        The UID of the SOP Class
    service_class_application_information : bytes
        The Service Class Application Information as per the Service Class
        Specifications (see PS3.4)

    Examples
    --------

    .. code-block:: python

       >>> from pynetdicom.pdu_items import SOPClassExtendedNegotiation
       >>> item = SOPClassExtendedNegotiation()
       >>> item.sop_class_uid = '1.2.840.10008.5.1.4.1.2.1.3'
       >>> item.service_class_application_information = b'\x01'

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.5
    """
    def __init__(self):
        self.sop_class_uid = None
        self.service_class_application_information = None

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.SOPClassExtendedNegotiationSubItem

        Raises
        ------
        ValueError
            If `sop_class_uid` or `service_class_application_information` are
            not set
        """
        if self.sop_class_uid is None \
                or self.service_class_application_information is None:
            LOGGER.error("SOP Class UID and Service Class Application "
                         "Information must be set prior to Association "
                         "negotiation")
            raise ValueError("SOP Class UID and Service Class Application "
                             "Information must be set prior to Association "
                             "negotiation")

        item = SOPClassExtendedNegotiationSubItem()
        item.from_primitive(self)

        return item

    @property
    def service_class_application_information(self):
        """Return the *Service Class Application Information*."""
        return self._service_class_application_information

    @service_class_application_information.setter
    def service_class_application_information(self, value):
        """Sets the Service Class Application Information parameter.

        Parameters
        ----------
        value : bytes
            The Service Class Application Information as per the Service Class
            Specifications (see PS3.4)

        Raises
        ------
        TypeError
            If `value` is not a bytes object
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("Service Class Application Information should be a "
                         "bytes object")
            raise TypeError("Service Class Application Information should "
                            "be a bytes object")

        self._service_class_application_information = value

    @property
    def sop_class_uid(self):
        """Return the *SOP Class UID*."""
        return self._sop_class_uid

    @sop_class_uid.setter
    def sop_class_uid(self, value):
        """Sets the SOP Class UID parameter.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The corresponding Abstract Syntax UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, UID):
            pass
        elif isinstance(value, str):
            value = UID(value)
        elif isinstance(value, bytes):
            value = UID(value.decode('ascii'))
        elif value is None:
            pass
        else:
            LOGGER.error("SOP Class UID must be a pydicom.uid.UID, str "
                         "or bytes")
            raise TypeError("SOP Class UID must be a pydicom.uid.UID, str "
                            "or bytes")

        if value is not None and not value.is_valid:
            LOGGER.error("Implementation Class UID is an invalid UID")
            raise ValueError("Implementation Class UID is an invalid UID")

        self._sop_class_uid = value


class SOPClassCommonExtendedNegotiation(ServiceParameter):
    """
    A representation of the SOP Class Common Extended Negotiation primitive.

    Allows peer AEs to exchange generic application information.

    The SOP Class Common Extended Negotiation is optional and there may only be
    a single SOPClassCommonExtendedNegotiation item for each available SOP
    Class UID.

    Identical for both A-ASSOCIATE-RQ and A-ASSOCIATE-AC

    Attributes
    ----------
    sop_class_uid : pydicom.uid.UID, bytes or str
        The UID of the SOP Class
    service_class_uid : pydicom.uid.UID, bytes or str
        The UID of the corresponding Service Class
    related_general_sop_class_identification : list of UID str
        Related General SOP Class UIDs (optional)

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import SOPClassCommonExtendedNegotiation
    >>> item = SOPClassCommonExtendedNegotiation()
    >>> item.sop_class_uid = '1.2.840.10008.5.1.4.1.1.88.40'
    >>> item.service_class_uid = '1.2.840.10008.4.2'
    >>> item.related_general_sop_class_identification = [
    ...     '1.2.840.10008.5.1.4.1.1.88.22'
    ... ]

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.6
    """
    def __init__(self):
        self.sop_class_uid = None
        self.service_class_uid = None
        self.related_general_sop_class_identification = []

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.SOPClassCommonExtendedNegotiationSubItem

        Raises
        ------
        ValueError
            If `sop_class_uid` or `service_class_uid` are not set
        """
        if self.sop_class_uid is None or self.service_class_uid is None:
            LOGGER.error("SOP Class UID and Service Class UID must be set "
                         "prior to Association negotiation")
            raise ValueError("SOP Class UID and Service Class UID must be "
                             "set prior to Association negotiation")

        item = SOPClassCommonExtendedNegotiationSubItem()
        item.from_primitive(self)

        return item

    @property
    def related_general_sop_class_identification(self):
        """Return the *Related General SOP Class Identification*."""
        return self._related_general_sop_class_identification

    @related_general_sop_class_identification.setter
    def related_general_sop_class_identification(self, uid_list):
        """Sets the Service Class Application Information parameter.

        Parameters
        ----------
        uid_list : list of (pydicom.uid.UID, bytes or str)
            A list containing UIDs to be used in the Related General SOP Class
            Identification parameter

        Raises
        ------
        TypeError
            If `uid_list` is not a list
        ValueError
            If `uid_list` contains items that aren't UIDs
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(uid_list, list):
            # Test that all the items in the list are UID compatible and
            #   convert them to pydicom.uid.UID if required
            valid_uid_list = []

            for uid in uid_list:
                if isinstance(uid, UID):
                    pass
                elif isinstance(uid, str):
                    uid = UID(uid)
                elif isinstance(uid, bytes):
                    uid = UID(uid.decode('ascii'))
                else:
                    LOGGER.error("Related General SOP Class Identification "
                                 "must be a list of pydicom.uid.UID, str "
                                 "or bytes")
                    raise TypeError("Related General SOP Class "
                                    "Identification must be a list of "
                                    "pydicom.uid.UID, str or bytes")

                if uid is not None and not uid.is_valid:
                    LOGGER.warning("Related General SOP Class "
                                   "Identification contains an invalid UID")

                valid_uid_list.append(uid)

            self._related_general_sop_class_identification = valid_uid_list
        else:
            LOGGER.error("Related General SOP Class Identification "
                         "must be a list of pydicom.uid.UID, str "
                         "or bytes")
            raise TypeError("Related General SOP Class Identification "
                            "must be a list of pydicom.uid.UID, str "
                            "or bytes")

    @property
    def service_class_uid(self):
        """Return the *Service Class UID*."""
        return self._service_class_uid

    @service_class_uid.setter
    def service_class_uid(self, value):
        """Sets the Service Class UID parameter.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The corresponding Service Class UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, UID):
            pass
        elif isinstance(value, str):
            value = UID(value)
        elif isinstance(value, bytes):
            value = UID(value.decode('ascii'))
        elif value is None:
            pass
        else:
            LOGGER.error("Service Class UID must be a pydicom.uid.UID, str "
                         "or bytes")
            raise TypeError("Service Class UID must be a pydicom.uid.UID, "
                            "str or bytes")

        if value is not None and not value.is_valid:
            LOGGER.warning("Implementation Class UID is an invalid UID")

        self._service_class_uid = value

    @property
    def sop_class_uid(self):
        """Return the *SOP Class UID*."""
        return self._sop_class_uid

    @sop_class_uid.setter
    def sop_class_uid(self, value):
        """Sets the SOP Class UID parameter.

        Parameters
        ----------
        value : pydicom.uid.UID, bytes or str
            The SOP Class UID

        Raises
        ------
        TypeError
            If `value` is not a pydicom.uid.UID, bytes or str
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, UID):
            pass
        elif isinstance(value, str):
            value = UID(value)
        elif isinstance(value, bytes):
            value = UID(value.decode('ascii'))
        elif value is None:
            pass
        else:
            LOGGER.error("SOP Class UID must be a pydicom.uid.UID, str "
                         "or bytes")
            raise TypeError("SOP Class UID must be a pydicom.uid.UID, str "
                            "or bytes")

        if value is not None and not value.is_valid:
            LOGGER.warning("Implementation Class UID is an invalid UID")

        self._sop_class_uid = value


class UserIdentityNegotiation(ServiceParameter):
    """Representation of the User Identity Negotiation primitive.

    Allows peer AEs to exchange generic application information.

    The SOP Class Common Extended Negotiation is optional and there may only be
    a single SOPClassCommonExtendedNegotiation item for each available SOP
    Class UID.

    In general, a User Identity Negotiation request that is accepted will
    result in Association establishment and possibly a server response if
    requested and supported by the peer. If a server response is requested but
    not received then the Requestor must decide how to proceed.
    An Association rejected due to an authorisation failure will be indicated
    using Rejection Permanent with a Source of "DICOM UL service provided (ACSE
    related function)".

    How the Acceptor handles authentication is to be implemented by the
    end-user and is outside the scope of the DICOM standard.

    A-ASSOCIATE-RQ
    | `user_identity_type`
    | `positive_response_requested`
    | `primary_field`
    | `secondary_field`

    A-ASSOCIATE-AC
    The `server_response` parameter is required when a response to the User
    Identity Negotiation request is to be issued (although this depends on
    whether or not this is supported by the Acceptor).

    Attributes
    ----------
    user_identity_type : int or None
        A-ASSOCIATE-RQ only. One of the following values:

        * 1 - Username as string in UTF-8
        * 2 - Username as string in UTF-8 and passcode
        * 3 - Kerberos Service ticket
        * 4 - SAML Assertion
        * 5 - JSON Web Token
    positive_response_requested : bool
        A-ASSOCIATE-RQ only. True when requesting a response, False otherwise
        (default is False)
    primary_field : bytes or None
        A-ASSOCIATE-RQ only. Contains either the username, Kerberos Service
        ticket or SAML assertion depending on `user_identity_type`.
    secondary_field : bytes or None
        A-ASSOCIATE-RQ only. Only required if the `user_identity_type` is 2,
        when it should contain the passcode as a bytes object, None otherwise
    server_response : bytes or None
        A-ASSOCIATE-AC only. Shall contain the Kerberos Service ticket or SAML
        response if the `user_identity_type` in the Request was 3 or 4. Shall
         be None if `user_identity_type` was 1 or 2.

    Examples
    --------

    >>> from pynetdicom.pdu_primitives import UserIdentityNegotiation
    >>> item = UserIdentityNegotiation()
    >>> item.user_identity_type = 2
    >>> item.positive_response_requested = True
    >>> item.primary_field = b'username'
    >>> item.secondary_field = b'password'

    References
    ----------

    * DICOM Standard, Part 7, Annex D.3.3.7
    """

    def __init__(self):
        self.user_identity_type = None
        self.positive_response_requested = False
        self.primary_field = None
        self.secondary_field = b''
        self.server_response = None

    def from_primitive(self):
        """Convert the primitive to a PDU item ready to be encoded.

        Returns
        -------
        item : pdu_items.UserIdentitySubItemRQ or
            pdu_items.UserIdentitySubItemAC

        Raises
        ------
        ValueError
            If server_response is None and user_identity_type or primary_field
            are None
        ValueError
            If server_response is None and user_identity_type is 2 and
            secondary_field is None
        """
        # Determine if this primitive is an -RQ or -AC
        if self.server_response is None:
            # Then an -RQ
            if self.user_identity_type is None or self.primary_field is None:
                msg = (
                    "User Identity Type and Primary Field must be "
                    "set prior to Association negotiation"
                )
                LOGGER.error(msg)
                raise ValueError(msg)

            if self.user_identity_type == 2 and self.secondary_field == b'':
                msg = "Secondary Field must be set when User Identity is 2"
                LOGGER.error(msg)
                raise ValueError(msg)

            item = UserIdentitySubItemRQ()

        else:
            # Then an -AC
            item = UserIdentitySubItemAC()

        item.from_primitive(self)

        return item

    @property
    def positive_response_requested(self):
        """Return *Positive Response Requested*."""
        return self._positive_response_requested

    @positive_response_requested.setter
    def positive_response_requested(self, value):
        """Sets the Positive Response Requested parameter.

        Parameters
        ----------
        value : bool
            True if response requested, False otherwise

        Raises
        ------
        TypeError
            If `value` is not a bool
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bool):
            pass
        else:
            LOGGER.error("Positive Response Requested must be boolean")
            raise TypeError("Positive Response Requested must be boolean")

        self._positive_response_requested = value

    @property
    def primary_field(self):
        """Return *Primary Field*."""
        return self._primary_field

    @primary_field.setter
    def primary_field(self, value):
        """Sets the Primary Field parameter.

        Parameters
        ----------
        value : bytes or None
            The username or Kerberos Service ticket as a bytes object

        Raises
        ------
        TypeError
            If `value` is not bytes or None
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("Primary Field must be bytes if requesting "
                         "Association, None otherwise")
            raise TypeError("Primary Field must be bytes if requesting "
                            "Association, None otherwise")

        self._primary_field = value

    @property
    def secondary_field(self):
        """Return the *Secondary Field*."""
        return self._secondary_field

    @secondary_field.setter
    def secondary_field(self, value):
        """Sets the Secondary Field parameter.

        Only used when User Identity Type is equal to 2.

        Parameters
        ----------
        value : bytes or None
            The passcode as a bytes object

        Raises
        ------
        TypeError
            If `value` is not bytes or None
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("Secondary Field must be bytes if requesting "
                         "Association with User Identity Type equal to 2, "
                         "None otherwise")
            raise TypeError("Secondary Field must be bytes if requesting "
                            "Association with User Identity Type equal to 2, "
                            "None otherwise")

        self._secondary_field = value

    @property
    def server_response(self):
        """Return the *Server Response*."""
        return self._server_response

    @server_response.setter
    def server_response(self, value):
        """Sets the Server Response parameter.

        Parameters
        ----------
        value : bytes or None
            The server response as a bytes object

        Raises
        ------
        TypeError
            If `value` is not bytes or None
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, bytes):
            pass
        elif value is None:
            pass
        else:
            LOGGER.error("Server Response must be bytes or None")
            raise TypeError("Server Response must be bytes or None")

        self._server_response = value

    def __str__(self):
        """String representation of the class."""
        s = 'User Identity Parameters\n'
        if self.server_response is None:
            s += '  User identity type: {0:d}\n'.format(
                self.user_identity_type)
            s += '  Positive response requested: {0!r}\n' \
                 .format(self.positive_response_requested)
            s += '  Primary field: {0!s}\n'.format(self.primary_field)
            s += '  Secondary field: {0!s}\n'.format(self.secondary_field)
        else:
            s += '  Server response: {0!s}\n'.format(self.server_response)

        return s

    @property
    def user_identity_type(self):
        """Return the *User Identity Type*."""
        return self._user_identity_type

    @user_identity_type.setter
    def user_identity_type(self, value):
        """Sets the User Identity Type parameter.

        Parameters
        ----------
        value : int
            One of the following:

            * 1 - Username as string in UTF-8
            * 2 - Username as string in UTF-8 and passcode
            * 3 - Kerberos Service ticket
            * 4 - SAML Assertion
            * 5 - JSON Web Token

        Raises
        ------
        TypeError
            If `value` is not an int or None
        ValueError
            If `value` is an int and is not 1, 2, 3 or 4
        """
        # pylint: disable=attribute-defined-outside-init
        if isinstance(value, int):
            if value not in [1, 2, 3, 4, 5]:
                LOGGER.error("User Identity Type must be 1, 2, 3, 4 or 5 if "
                             "requesting Association, None otherwise")
                raise ValueError("User Identity Type must be 1, 2, 3, 4 or 5 "
                                 "if requesting Association, None otherwise")
        elif value is None:
            pass
        else:
            LOGGER.error("User Identity Type must be an int or None")
            raise TypeError("User Identity Type must be an int or None")

        self._user_identity_type = value



"""Tests for the NonPatientObjectStorageServiceClass."""

from io import BytesIO
import logging
import os
import threading
import time

import pytest

from pydicom import dcmread
from pydicom.dataset import Dataset
from pydicom.uid import ExplicitVRLittleEndian

from pynetdicom import AE
from pynetdicom.dimse_primitives import C_STORE
from pynetdicom.sop_class import (
    NonPatientObjectStorageServiceClass,
    HangingProtocolStorage,
)
from .dummy_c_scp import (
    DummyBaseSCP,
    DummyStorageSCP
)

LOGGER = logging.getLogger('pynetdicom')
#LOGGER.setLevel(logging.DEBUG)
LOGGER.setLevel(logging.CRITICAL)

TEST_DS_DIR = os.path.join(os.path.dirname(__file__), 'dicom_files')
DATASET = dcmread(os.path.join(TEST_DS_DIR, 'CTImageStorage.dcm'))
DATASET.SOPClassUID = HangingProtocolStorage


class TestNonPatientObjectStorageServiceClass(object):
    """Test the NonPatientObjectStorageServiceClass"""
    def setup(self):
        """Run prior to each test"""
        self.scp = None

    def teardown(self):
        """Clear any active threads"""
        if self.scp:
            self.scp.abort()

        time.sleep(0.1)

        for thread in threading.enumerate():
            if isinstance(thread, DummyBaseSCP):
                thread.abort()
                thread.stop()

    @pytest.mark.skip("Not aware of any way to test")
    def test_scp_failed_ds_decode(self):
        """Test failure to decode the dataset"""
        # Hard to test directly as decode errors won't show up until the
        #   dataset is actually used
        self.scp = DummyStorageSCP()
        self.scp.status = 0x0000
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage, ExplicitVRLittleEndian)
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established

        req = C_STORE()
        req.MessageID = 1
        req.AffectedSOPClassUID = DATASET.SOPClassUID
        req.AffectedSOPInstanceUID = DATASET.SOPInstanceUID
        req.Priorty = 0x0002
        req.DataSet = BytesIO(b'\x08\x00\x01\x00\x04\x00\x00\x00\x00\x08\x00\x49')

        # Send C-STORE request to DIMSE and get response
        assoc.dimse.send_msg(req, 1)
        rsp, _ = assoc.dimse.receive_msg(True)

        assert rsp.Status == 0xC100
        assert rsp.ErrorComment == 'Unable to decode the dataset'
        assoc.release()
        self.scp.stop()

    def test_scp_callback_return_dataset(self):
        """Test on_c_store returning a Dataset status"""
        self.scp = DummyStorageSCP()
        self.scp.status = Dataset()
        self.scp.status.Status = 0x0001
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0x0001
        assoc.release()
        self.scp.stop()

    def test_scp_callback_return_dataset_multi(self):
        """Test on_c_store returning a Dataset status with other elements"""
        self.scp = DummyStorageSCP()
        self.scp.status = Dataset()
        self.scp.status.Status = 0x0001
        self.scp.status.ErrorComment = 'Test'
        self.scp.status.OffendingElement = 0x00080010
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0x0001
        assert rsp.ErrorComment == 'Test'
        assert rsp.OffendingElement == 0x00080010
        assoc.release()
        self.scp.stop()

    def test_scp_callback_return_int(self):
        """Test on_c_echo returning an int status"""
        self.scp = DummyStorageSCP()
        self.scp.status = 0x0000
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0x0000
        assert not 'ErrorComment' in rsp
        assoc.release()
        self.scp.stop()

    def test_scp_callback_return_invalid(self):
        """Test on_c_store returning a valid status"""
        self.scp = DummyStorageSCP()
        self.scp.status = 0xFFF0
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0xFFF0
        assoc.release()
        self.scp.stop()

    def test_scp_callback_no_status(self):
        """Test on_c_store not returning a status"""
        self.scp = DummyStorageSCP()
        self.scp.status = None
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0xC002
        assoc.release()
        self.scp.stop()

    def test_scp_callback_exception(self):
        """Test on_c_store raising an exception"""
        self.scp = DummyStorageSCP()
        def on_c_store(ds, context, assoc_info):
            raise ValueError
        self.scp.ae.on_c_store = on_c_store
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        rsp = assoc.send_c_store(DATASET)
        assert rsp.Status == 0xC211
        assoc.release()
        self.scp.stop()

    def test_scp_callback_context(self):
        """Test on_c_store caontext parameter"""
        self.scp = DummyStorageSCP()
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage, '1.2.840.10008.1.2.1')
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET)
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        assert self.scp.context.context_id == 1
        assert self.scp.context.abstract_syntax == HangingProtocolStorage
        assert self.scp.context.transfer_syntax == '1.2.840.10008.1.2.1'

        self.scp.stop()

    def test_scp_callback_info(self):
        """Test on_c_store caontext parameter"""
        self.scp = DummyStorageSCP()
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET)
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        assert 'address' in self.scp.info['requestor']
        assert self.scp.info['requestor']['ae_title'] == b'PYNETDICOM      '
        #assert self.scp.info['requestor']['called_aet'] == b'ANY-SCP         '
        assert isinstance(self.scp.info['requestor']['port'], int)
        assert self.scp.info['acceptor']['port'] == 11112
        assert 'address' in self.scp.info['acceptor']
        assert self.scp.info['acceptor']['ae_title'] == b'PYNETDICOM      '
        assert self.scp.info['parameters']['message_id'] == 1
        assert self.scp.info['parameters']['priority'] == 2
        assert self.scp.info['parameters']['originator_aet'] is None
        assert self.scp.info['parameters']['originator_message_id'] is None

        self.scp.stop()

    def test_scp_callback_info_move_origin(self):
        """Test on_c_store caontext parameter"""
        self.scp = DummyStorageSCP()
        self.scp.start()

        ae = AE()
        ae.add_requested_context(HangingProtocolStorage)
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate('localhost', 11112)
        assert assoc.is_established
        status = assoc.send_c_store(DATASET,
                                    originator_aet=b'ORIGIN',
                                    originator_id=888)
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        assert 'address' in self.scp.info['requestor']
        assert self.scp.info['requestor']['ae_title'] == b'PYNETDICOM      '
        #assert self.scp.info['requestor']['called_aet'] == b'ANY-SCP         '
        assert isinstance(self.scp.info['requestor']['port'], int)
        assert self.scp.info['acceptor']['port'] == 11112
        assert 'address' in self.scp.info['acceptor']
        assert self.scp.info['acceptor']['ae_title'] == b'PYNETDICOM      '
        assert self.scp.info['parameters']['message_id'] == 1
        assert self.scp.info['parameters']['priority'] == 2
        assert self.scp.info['parameters']['originator_aet'] == b'ORIGIN          '
        assert self.scp.info['parameters']['originator_message_id'] == 888

        self.scp.stop()



"""Implementation of the DIMSE Status values."""

from pydicom.dataset import Dataset

from pynetdicom._globals import (
    STATUS_SUCCESS,
    STATUS_FAILURE,
    STATUS_WARNING,
    STATUS_CANCEL,
    STATUS_PENDING,
    STATUS_UNKNOWN,
)


# Non-Service Class specific statuses - PS3.7 Annex C
GENERAL_STATUS = {
    0x0000 : (STATUS_SUCCESS, ''),
    0x0105 : (STATUS_FAILURE, 'No Such Attribute'),
    0x0106 : (STATUS_FAILURE, 'Invalid Attribute Value'),
    0x0107 : (STATUS_WARNING, 'Attribute List Error'),
    0x0110 : (STATUS_FAILURE, 'Processing Failure'),
    0x0111 : (STATUS_FAILURE, 'Duplication SOP Instance'),
    0x0112 : (STATUS_FAILURE, 'No Such SOP Instance'),
    0x0113 : (STATUS_FAILURE, 'No Such Event Type'),
    0x0114 : (STATUS_FAILURE, 'No Such Argument'),
    0x0115 : (STATUS_FAILURE, 'Invalid Argument Value'),
    0x0116 : (STATUS_WARNING, 'Attribute Value Out of Range'),
    0x0117 : (STATUS_FAILURE, 'Invalid Object Instance'),
    0x0118 : (STATUS_FAILURE, 'No Such SOP Class'),
    0x0119 : (STATUS_FAILURE, 'Class-Instance Conflict'),
    0x0120 : (STATUS_FAILURE, 'Missing Attribute'),
    0x0121 : (STATUS_FAILURE, 'Missing Attribute Value'),
    0x0122 : (STATUS_FAILURE, 'Refused: SOP Class Not Supported'),
    0x0123 : (STATUS_FAILURE, 'No Such Action'),
    0x0124 : (STATUS_FAILURE, 'Refused: Not Authorised'),
    0x0210 : (STATUS_FAILURE, 'Duplicate Invocation'),
    0x0211 : (STATUS_FAILURE, 'Unrecognised Operation'),
    0x0212 : (STATUS_FAILURE, 'Mistyped Argument'),
    0x0213 : (STATUS_FAILURE, 'Resource Limitation'),
    0xFE00 : (STATUS_CANCEL, '')
}


## SERVICE CLASS STATUSES
# Verification Service Class specific status code values
VERIFICATION_SERVICE_CLASS_STATUS = GENERAL_STATUS


# Storage Service Class specific status code values - PS3.4 Annex B.2.3
STORAGE_SERVICE_CLASS_STATUS = {
    0xB000 : (STATUS_WARNING, 'Coercion of Data Elements'),
    0xB007 : (STATUS_WARNING, 'Data Set Does Not Match SOP Class'),
    0xB006 : (STATUS_WARNING, 'Element Discarded')
}

# Ranged values
for _code in range(0xA700, 0xA7FF + 1):
    STORAGE_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE,
                                           'Refused: Out of Resources')
for _code in range(0xA900, 0xA9FF + 1):
    STORAGE_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE,
                                           'Data Set Does Not Match SOP Class')
for _code in range(0xC000, 0xCFFF + 1):
    STORAGE_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, 'Cannot Understand')

# Add the General status code values - PS3.7 9.1.1.1.9 and Annex C
STORAGE_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Query Retrieve - FIND specific status code values
#   PS3.4 Annex C.4.1.1.4
# Hanging Protocol - FIND specific status code values
#   PS3.4 Annex U.4.1
# Color Palette - FIND specific status code values
#   PS3.4 Annex X
# Implant Template - FIND specific status code values
#   PS3.4 Annex BB
# Defined Procedure Protocol - FIND specific status code values
#   PS3.4 Annex HH
QR_FIND_SERVICE_CLASS_STATUS = {
    0xA700 : (STATUS_FAILURE, 'Refused: Out of Resources'),
    0xA900 : (STATUS_FAILURE, 'Identifier Does Not Match SOP Class'),
    0xFF00 : (STATUS_PENDING,
              'Matches are continuing, current match supplied'),
    0xFF01 : (STATUS_PENDING, 'Matches are continuing, warning')
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    QR_FIND_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, 'Unable to Process')

# Add the General status code values - PS3.7 Annex C
QR_FIND_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Query Retrieve - MOVE specific status code values
#   PS3.4 Annex C.4.2.1.5 and Annex Y.4.1.1.4
# Hanging Protocol - MOVE specific status code values
#   PS3.4 Annex U.4.2
# Color Palette - MOVE specific status code values
#   PS3.4 Annex X
# Implant Template - MOVE specific status code values
#   PS3.4 Annex BB
# Defined Procedure Protocol - MOVE specific status code values
#   PS3.4 Annex HH
QR_MOVE_SERVICE_CLASS_STATUS = {
    0xA701 : (STATUS_FAILURE,
              'Refused: Out of resources, unable to calculate '
              'number of matches'),
    0xA702 : (STATUS_FAILURE,
              'Refused: Out of resources, unable to perform sub-operations'),
    0xA801 : (STATUS_FAILURE, 'Move destination unknown'),
    0xA900 : (STATUS_FAILURE, 'Identifier does not match SOP class'),
    0xAA00 : (STATUS_FAILURE,
              "None of the frames requested were found in the SOP instance"),
    0xAA01 : (STATUS_FAILURE,
              "Unable to create new object for this SOP class"),
    0xAA02 : (STATUS_FAILURE, "Unable to extract frames"),
    0xAA03 : (STATUS_FAILURE,
              "Time-based request received for a non-time-based original "
              "SOP Instance"),
    0xAA04 : (STATUS_FAILURE, "Invalid request"),
    0xFF00 : (STATUS_PENDING, 'Sub-operations are continuing'),
    0xB000 : (STATUS_WARNING, 'Sub-operations completed, one or more failures')
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    QR_MOVE_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, 'Unable to Process')

# Add the General status code values - PS3.7 Annex C
QR_MOVE_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Query Retrieve - GET specific status code values
#   PS3.4 Annex C.4.3.1.4 and Annex Y.4.2.1.4
# Hanging Protocol  - GET specific status code values
#   PS3.4 Annex U.4.3
# Color Palette - GET specific status code values
#   PS3.4 Annex X
# Implant Template - GET specific status code values
#   PS3.4 Annex BB
# Defined Procedure Protocol - GET specific status code values
#   PS3.4 Annex HH
QR_GET_SERVICE_CLASS_STATUS = {
    0xA701 : (STATUS_FAILURE,
              'Refused: Out of resources, unable to calculate '
              'number of matches'),
    0xA702 : (STATUS_FAILURE,
              'Refused: Out of resources, unable to perform sub-operations'),
    0xA900 : (STATUS_FAILURE, 'Identifier does not match SOP class'),
    0xAA00 : (STATUS_FAILURE,
              "None of the frames requested were found in the SOP instance"),
    0xAA01 : (STATUS_FAILURE,
              "Unable to create new object for this SOP class"),
    0xAA02 : (STATUS_FAILURE, "Unable to extract frames"),
    0xAA03 : (STATUS_FAILURE,
              "Time-based request received for a non-time-based original "
              "SOP Instance"),
    0xAA04 : (STATUS_FAILURE, "Invalid request"),
    0xFF00 : (STATUS_PENDING, 'Sub-operations are continuing'),
    0xB000 : (STATUS_WARNING,
              'Sub-operations complete, one or more failures or warnings'),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    QR_GET_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE, 'Unable to Process')

# Add the General status code values - PS3.7 Annex C
QR_GET_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Modality Worklist Service Class specific status code values
#   PS3.4 Annex K
MODALITY_WORKLIST_SERVICE_CLASS_STATUS = {
    0xA700 : (STATUS_FAILURE, 'Refused: Out of resources'),
    0xA900 : (STATUS_FAILURE, 'Identifier does not match SOP class'),
    0xFF00 : (STATUS_PENDING,
              "Matches are continuing - current match is supplied and any "
              "Optional Keys were supported in the same manner as Required "
              "Keys"),
    0xFE00 : (STATUS_PENDING,
              "Matches are continuing - warning that one or more Optional "
              "Keys were not supported for existence for this Identifier"),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    MODALITY_WORKLIST_SERVICE_CLASS_STATUS[_code] = (STATUS_FAILURE,
                                                     'Unable to Process')

# Add the General status code values - PS3.7 Annex C
MODALITY_WORKLIST_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Relevant Patient Information Query Service Class specific status code values
# PS3.4 Annex Q
RELEVANT_PATIENT_SERVICE_CLASS_STATUS = {
    0xA700 : (STATUS_FAILURE, "Out of resources"),
    0xA900 : (STATUS_FAILURE, "Identifier doesn't match SOP Class"),
    0xC000 : (STATUS_FAILURE, "Unable to process"),
    0xC100 : (STATUS_FAILURE, "More than one match found"),
    0xC200 : (STATUS_FAILURE, "Unable to support requested template"),
    0xFF00 : (STATUS_PENDING, "Current match is supplied"),
}
RELEVANT_PATIENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Substance Administration Query Service Class specific status code values
#   PS3.4 Annex V
SUBSTANCE_ADMINISTRATION_SERVICE_CLASS_STATUS = {
    0xA700 : (STATUS_FAILURE, "Out of resources"),
    0xA900 : (STATUS_FAILURE, "Data set doesn't match SOP Class"),
    0xFF00 : (STATUS_PENDING,
              "Matches are continuing, current match is supplied and any "
              "Optional Keys were supported in the same manner as Required "
              "Keys"),
    0xFF01 : (STATUS_PENDING,
              "Matches are continuing, warning that one or more Optional "
              "Keys were not supported for existence for this Identifier")
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    SUBSTANCE_ADMINISTRATION_SERVICE_CLASS_STATUS[_code] = (
        STATUS_FAILURE, 'Unable to Process'
    )

SUBSTANCE_ADMINISTRATION_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Non-Patient Object Storage Service Class specific status code values
#   PS3.4 Annex GG
NON_PATIENT_SERVICE_CLASS_STATUS = {
    0xA700 : (STATUS_FAILURE, "Out of resources"),
    0xA900 : (STATUS_FAILURE, "Data set doesn't match SOP Class"),
    0xC000 : (STATUS_FAILURE, "Cannot understand"),
}
NON_PATIENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Procedure Step SOP Class specific status code values - WIP
#   PS3.4 Annex F
_PROCEDURE_STEP_STATUS = {
    0x0001 : (STATUS_WARNING,
              "Requested optional attributes are not supported"),
    0x0110 : (STATUS_FAILURE,
              "Performed Procedure Step object may no longer be updated"),
}
_PROCEDURE_STEP_STATUS.update(GENERAL_STATUS)


# Print Job Management Service Class specific status code values - WIP
#   PS3.4 Annex H
_PRINT_JOB_MANAGEMENT_SERVICE_CLASS_STATUS = {
    0xB600 : (STATUS_WARNING, "Memory allocation not supported"),
    0xB601 : (STATUS_WARNING,
              "Film session printing (collation) is not supported"),
    0xB602 : (STATUS_WARNING,
              "Film Session SOP Instance hierarchy does not contain Image "
              "Box SOP Instances (empty page)"),
    0xB603 : (STATUS_WARNING,
              "Film Box SOP Instance hierarchy does not contain Image Box "
              "SOP Instances (empty page)"),
    0xB604 : (STATUS_WARNING,
              "Image size is larger than image box size, the image has been "
              "demagnified"),
    0xB605 : (STATUS_WARNING,
              "Requested minimum density or maximum density outside of "
              "printer's operating range. The print will use its respective "
              "minimum or maximum density value instead"),
    0xB609 : (STATUS_WARNING,
              "Image size is larger than the image box size, the image has "
              "been cropped to fit"),
    0xB60A : (STATUS_WARNING,
              "Image size or Combined Print Image size is larger than the "
              "image box size, image or combined print image has been "
              "decimated to fit"),
    0xC600 : (STATUS_FAILURE,
              "Film Session SOP Instance hierarchy does not contain Film Box "
              "SOP Instances"),
    0xC601 : (STATUS_FAILURE,
              "Unable to create Print Job SOP Instance; print queue is full"),
    0xC602 : (STATUS_FAILURE,
              "Unable to create Print Job SOP instance; print queue is full"),
    0xC603 : (STATUS_FAILURE, "Image size is larger than image box size"),
    0xC605 : (STATUS_FAILURE,
              "Insufficient memory in printer to store the image"),
    0xC613 : (STATUS_FAILURE,
              "Combined print image size is larger than the image box size"),
    0xC616 : (STATUS_FAILURE,
              "There is an existing film box that has not been printed and "
              "N-ACTION at the film session level is not supported. A new "
              "film box will not be created when a previous film box has not "
              "been printed"),
}
_PRINT_JOB_MANAGEMENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Storage Commitment Service Class specific status code values - WIP
#   PS3.4 Annex J
_STORAGE_COMMITMENT_SERVICE_CLASS_STATUS = GENERAL_STATUS


# Application Event Logging Service Class specific status code values - WIP
#   PS3.4 Annex N
_APPLICATION_EVENT_LOGGING_SERVICE_CLASS_STATUS = {
    0xB101 : (STATUS_WARNING,
              "Specified Synchronisation Frame of Reference UID doesn't "
              "match SCP Synchronization Frame of Reference"),
    0xB102 : (STATUS_WARNING,
              "Study Instance UID coercion; event logged under a different "
              "Study Instance UID"),
    0xB104 : (STATUS_WARNING,
              "IDs inconsistent in matching a current study; event logged"),
    0xC101 : (STATUS_FAILURE,
              "Procedural logging not available for specified Study "
              "Instance UID"),
    0xC102 : (STATUS_FAILURE, "Event Information doesn't match template"),
    0xC103 : (STATUS_FAILURE, "Cannot match event to a current study"),
    0xC104 : (STATUS_FAILURE,
              "IDs inconsistent in matching a current study; event not "
              "logged"),
    0xC10E : (STATUS_FAILURE,
              "Operator not authorised to add entry to Medication "
              "Administration Record"),
    0xC110 : (STATUS_FAILURE,
              "Patient cannot be identified from Patient ID (0010,0020) or "
              "Admission ID (0038,0010)"),
    0xC111 : (STATUS_FAILURE,
              "Update of Medication Administration Record failed"),
}
_APPLICATION_EVENT_LOGGING_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Media Creation Management Service Class specific status code values - WIP
_MEDIA_CREATION_MANAGEMENT_SERVICE_CLASS_STATUS = {
    0x0001 : (STATUS_WARNING,
              "Requested optional attributes are not supported"),
    0xA510 : (STATUS_FAILURE,
              "Refused because an Initiate Media Creation action has already "
              "been received for this SOP Instance"),
    0xC201 : (STATUS_FAILURE, "Media creation request already completed"),
    0xC202 : (STATUS_FAILURE,
              "Media creation request already in progress and cannot be "
              "interrupted"),
    0xC203 : (STATUS_FAILURE, "Cancellation denied for unspecified reason"),
}
_MEDIA_CREATION_MANAGEMENT_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# Unified Procedure Step Service specific status code values - WIP
_UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS = {
    0x0001 : (STATUS_WARNING,
              "Requested optional attributes are not supported"),
    0xA700 : (STATUS_FAILURE, "Out of resources"),
    0xA900 : (STATUS_FAILURE, "Identifier doesn't match SOP Class"),
    0xB300 : (STATUS_WARNING, "The UPS was created with modifications"),
    0xB304 : (STATUS_WARNING,
              "The UPS is already in the requested state of CANCELED"),
    0xB305 : (STATUS_WARNING, "Coerced invalid values to valid values"),
    0xB306 : (STATUS_WARNING,
              "The UPS is already in the requested state of COMPLETED"),
    0xC300 : (STATUS_FAILURE, "The UPS may no longer be updated"),
    0xC301 : (STATUS_FAILURE, "The correct Transaction UID was not provided"),
    0xC302 : (STATUS_FAILURE, "The UPS is already IN PROGRESS"),
    0xC303 : (STATUS_FAILURE,
              "The UPS may only become SCHEDULED via N-CREATE, not N-SET or "
              "N-ACTION"),
    0xC304 : (STATUS_FAILURE,
              "The UPS has not met final state requirements for the "
              "requested state change"),
    0xC307 : (STATUS_FAILURE,
              "Specified SOP Instance UID does not exist or is not a UPS "
              "Instance managed by this SCP"),
    0xC308 : (STATUS_FAILURE, "Receiving AE-TITLE is unknown to this SCP"),
    0xC309 : (STATUS_FAILURE,
              "The provided value of UPS State was not SCHEDULED"),
    0xC310 : (STATUS_FAILURE, "The UPS is not yet in the IN PROGRESS state"),
    0xC311 : (STATUS_FAILURE, "The UPS is already COMPLETED"),
    0xC312 : (STATUS_FAILURE, "The performer cannot be contacted"),
    0xC313 : (STATUS_FAILURE, "Performer chooses not to cancel"),
    0xC314 : (STATUS_FAILURE,
              "Specified action is not appropriate for specified instance"),
    0xC315 : (STATUS_FAILURE, "SCP does not support Event Reports"),
    0xFF00 : (STATUS_PENDING,
              "Matches are continuing - current match is supplied an any "
              "Optional Keys were supported in the same manner as Required "
              "Keys"),
    0xFF01 : (STATUS_PENDING,
              "Matches are continuing - current match is supplied an any "
              "Optional Keys were not supported for existence for this "
              "Identifier"),
}

# Ranged values
for _code in range(0xC000, 0xCFFF + 1):
    _UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS[_code] = (
        STATUS_FAILURE, 'Unable to Process'
    )

_UNIFIED_PROCEDURE_STEP_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


# RT Machine Verification Service Class specific status code values - WIP
_RT_MACHINE_VERIFICATION_SERVICE_CLASS_STATUS = {
    0xC112 : (STATUS_FAILURE,
              "No such object instance - applicable Machine Verification "
              "instance not found"),
    0xC221 : (STATUS_FAILURE,
              "The Referenced Fraction Group Number does not exist in "
              "the referenced plan"),
    0xC222 : (STATUS_FAILURE,
              "No beams exist within the referenced fraction group"),
    0xC223 : (STATUS_FAILURE,
              "SCU already verifying and cannot currently process this "
              "request"),
    0xC224 : (STATUS_FAILURE,
              "Referenced Beam Number not found within the referenced "
              "Fraction Group"),
    0xC225 : (STATUS_FAILURE,
              "Referenced device or accessory not supported"),
    0xC226 : (STATUS_FAILURE,
              "Referenced device or accessory not found within the "
              "referenced beam"),
    0xC227 : (STATUS_FAILURE,
              "No such object instance - Referenced RT Plan not found"),
}
_RT_MACHINE_VERIFICATION_SERVICE_CLASS_STATUS.update(GENERAL_STATUS)


def code_to_status(code):
    """Return a Dataset with a Status element value of `code`."""
    if isinstance(code, int) and code >= 0:
        ds = Dataset()
        ds.Status = code
        return ds
    else:
        raise ValueError("'code' must be a positive integer.")


def code_to_category(code):
    """Return a Status' category as a str or 'Unknown' if not recognised.

    References
    ----------
    DICOM Standard Part 7, Annex C
    DICOM Standard Part 4
    """
    # pylint: disable=too-many-return-statements
    if isinstance(code, int) and code >= 0:
        if code == 0x0000:
            return STATUS_SUCCESS
        elif code in [0xFF00, 0xFF01]:
            return STATUS_PENDING
        elif code == 0xFE00:
            return STATUS_CANCEL
        elif code in [0x0105, 0x0106, 0x0110, 0x0111, 0x0112, 0x0113, 0x0114,
                      0x0115, 0x0117, 0x0118, 0x0119, 0x0120, 0x0121, 0x0122,
                      0x0123, 0x0124, 0x0210, 0x0211, 0x0212, 0x0213]:
            return STATUS_FAILURE
        elif code in range(0xA000, 0xB000):
            return STATUS_FAILURE
        elif code in range(0xC000, 0xD000):
            return STATUS_FAILURE
        elif code in [0x0107, 0x0116]:
            return STATUS_WARNING
        elif code in range(0xB000, 0xC000):
            return STATUS_WARNING
        elif code == 0x0001:
            return STATUS_WARNING

        return STATUS_UNKNOWN
    else:
        raise ValueError("'code' must be a positive integer.")



"""Unit tests for the Timer class."""

import logging
import time

import pytest

from pynetdicom.timer import Timer

LOGGER = logging.getLogger('pynetdicom')
LOGGER.setLevel(logging.CRITICAL)


class TestTimer(object):
    """Test the Timer class."""
    def test_init(self):
        """Test Timer initialisation"""
        timer = Timer(10)
        assert timer.timeout == 10
        timer = Timer(0)
        assert timer.timeout == 0
        timer = Timer(None)
        assert timer.timeout is None

    def test_property_setters_getters(self):
        """Test Timer property setters and getters."""
        timer = Timer(0)
        assert timer.timeout == 0
        assert not timer.expired
        assert timer.remaining == 0

        timer = Timer(None)
        assert timer.timeout is None
        assert not timer.expired
        assert timer.remaining == 1

        timer.timeout = 10
        assert timer.timeout == 10
        assert not timer.expired
        assert timer.remaining == 10

        timer.timeout = 0.2
        timer.start()
        time.sleep(0.1)
        assert timer.remaining < 0.1
        assert not timer.expired
        time.sleep(0.1)
        assert timer.expired

        timer.timeout = None
        assert timer.timeout is None

    def test_start_stop(self):
        """Test Timer stops."""
        timer = Timer(0.2)
        timer.start()
        time.sleep(0.1)
        timer.stop()
        time.sleep(0.2)
        assert not timer.expired

    def test_restart(self):
        """Test Timer restarts correctly."""
        timer = Timer(0.2)
        timer.start()
        time.sleep(0.1)
        assert timer.expired is False
        timer.restart()
        time.sleep(0.15)
        assert timer.expired is False
        time.sleep(0.1)
        assert timer.expired is True

    def test_no_timeout(self):
        """Test the timer with no time out."""
        timer = Timer(None)
        assert timer.timeout is None
        assert timer.expired is False
        assert timer.remaining is 1
        timer.start()
        assert timer.expired is False
        assert timer.remaining is 1
        time.sleep(0.5)
        assert timer.expired is False
        assert timer.remaining is 1
        timer.stop()
        assert timer.expired is False
        assert timer.remaining is 1

    def test_timeout(self):
        """Test the timer with a time out."""
        timer = Timer(0.1)
        assert timer.timeout == 0.1
        assert timer.expired is False
        assert timer.remaining == 0.1
        timer.start()
        assert timer.expired is False
        assert timer.remaining > 0
        time.sleep(0.2)
        assert timer.expired is True
        assert timer.remaining < 0
        timer.stop()
        assert timer.expired is True
        assert timer.remaining < 0

    def test_timeout_stop(self):
        """Test stopping the timer."""
        timer = Timer(0.1)
        assert timer.timeout == 0.1
        assert timer.expired is False
        assert timer.remaining == 0.1
        timer.start()
        timer.stop()
        assert timer.timeout == 0.1
        assert timer.expired is False
        assert timer.remaining > 0
        timer.start()
        time.sleep(0.2)
        timer.stop()
        assert timer.timeout == 0.1
        assert timer.expired is True
        assert timer.remaining < 0



"""Implements the supported Service Classes that make use of DIMSE-N."""

from io import BytesIO
import logging

from pynetdicom.dsutils import encode
from pynetdicom.dimse_primitives import N_GET
from pynetdicom.service_class import ServiceClass
from pynetdicom.status import GENERAL_STATUS, code_to_category
from pynetdicom._globals import (
    #STATUS_FAILURE,
    STATUS_SUCCESS,
    STATUS_WARNING,
    #STATUS_PENDING,
    #STATUS_CANCEL,
)


LOGGER = logging.getLogger('pynetdicom.service-n')


class DisplaySystemManagementServiceClass(ServiceClass):
    """Implementation of the Display System Management Service Class."""
    statuses = GENERAL_STATUS

    def SCP(self, req, context, info):
        """The implementation for the DIMSE N-GET service.

        Parameters
        ----------
        req : dimse_primitives.C_ECHO
            The N-GET request primitive sent by the peer.
        context : presentation.PresentationContext
            The presentation context that the service is operating under.
        info : dict
            A dict containing details about the association.

        See Also
        --------
        ae.ApplicationEntity.on_n_get
        association.Association.send_n_get

        Notes
        -----
        **N-GET Request**

        *Parameters*

        | (M) Message ID
        | (M) Requested SOP Class UID
        | (M) Requested SOP Instance UID
        | (U) Attribute Identifier List

        *Attribute Identifier List*

        An element with VR AT, VM 1-n, containing an attribute tag for each
        of the attributes applicable to the N-GET operation.

        **N-GET Response**

        *Parameters*

        | (M) Message ID Being Responded To
        | (U) Affected SOP Class UID
        | (U) Affected SOP Instance UID
        | (C) Attribute List
        | (M) Status

        *Attribute List*

        A dataset containing the values of the requested attributes.

        References
        ----------

        * DICOM Standard, Part 4, `Annex EE <http://dicom.nema.org/medical/dicom/current/output/html/part04.html#chapter_EE>`_
        * DICOM Standard, Part 7, Sections
          `10.1.2 <http://dicom.nema.org/medical/dicom/current/output/html/part07.html#sect_10.1.2>`_,
          `10.3.2 <http://dicom.nema.org/medical/dicom/current/output/html/part07.html#sect_10.3.2>`_
          and `Annex C <http://dicom.nema.org/medical/dicom/current/output/html/part07.html#chapter_C>`_
        """
        # Build N-GET response primitive
        rsp = N_GET()
        rsp.MessageIDBeingRespondedTo = req.MessageID
        rsp.AffectedSOPClassUID = req.RequestedSOPClassUID
        rsp.AffectedSOPInstanceUID = req.RequestedSOPInstanceUID

        info['parameters'] = {
            'message_id' : req.MessageID,
            'requested_sop_class_uid' : req.RequestedSOPClassUID,
            'requested_sop_instance_uid' : req.RequestedSOPInstanceUID,
        }

        # Attempt to run the ApplicationEntity's on_n_get callback
        # pylint: disable=broad-except
        try:
            # Send the value rather than the element
            (rsp_status, ds) = self.ae.on_n_get(req.AttributeIdentifierList,
                                                context.as_tuple, info)
        except Exception as exc:
            LOGGER.error(
                "Exception in the ApplicationEntity.on_n_get() callback"
            )
            LOGGER.exception(exc)
            # Processing failure - Error in on_n_get callback
            rsp_status = 0x0110

        # Validate rsp_status and set rsp.Status accordingly
        rsp = self.validate_status(rsp_status, rsp)

        # Success or Warning, must return AttributeList dataset
        if code_to_category(rsp.Status) in [STATUS_SUCCESS, STATUS_WARNING]:
            # Encode the `dataset` using the agreed transfer syntax
            #   Will return None if failed to encode
            transfer_syntax = context.transfer_syntax[0]
            bytestream = encode(ds,
                                transfer_syntax.is_implicit_VR,
                                transfer_syntax.is_little_endian)

            if bytestream is not None:
                rsp.AttributeList = BytesIO(bytestream)
            else:
                LOGGER.error("Failed to encode the supplied Dataset")
                # Processing failure - Failed to encode dataset
                rsp.Status = 0x0110

        self.dimse.send_msg(rsp, context.context_id)



"""
DICOM Dataset utility functions.
"""
import logging

from pydicom.filebase import DicomBytesIO
from pydicom.filereader import read_dataset
from pydicom.filewriter import write_dataset, write_data_element

LOGGER = logging.getLogger('pynetdicom.dsutils')

def decode(bytestring, is_implicit_vr, is_little_endian):
    """Decode `bytestring` to a pydicom Dataset.

    Parameters
    ----------
    byestring : io.BytesIO
        The encoded dataset in the DIMSE Message sent from the peer AE.
    is_implicit_vr : bool
        The dataset is encoded as implicit or explicit VR.
    is_little_endian : bool
        The byte ordering of the encoded dataset, little or big endian.

    Returns
    -------
    pydicom.dataset.Dataset
        The decoded dataset.
    """
    ## Logging
    transfer_syntax = "Little Endian" if is_little_endian else "Big Endian"
    if is_implicit_vr:
        transfer_syntax += " Implicit"
    else:
        transfer_syntax += " Explicit"

    LOGGER.debug('pydicom.read_dataset() TransferSyntax="%s"', transfer_syntax)

    ## Decode the dataset
    # Rewind to the start of the stream
    bytestring.seek(0)
    return read_dataset(bytestring, is_implicit_vr, is_little_endian)

def encode(ds, is_implicit_vr, is_little_endian):
    """Encode a pydicom Dataset `ds` to bytes.

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The dataset to encode
    is_implicit_vr : bool
        The element encoding scheme the dataset will be encoded with.
    is_little_endian : bool
        The byte ordering the dataset will be encoded in.

    Returns
    -------
    bytes or None
        The encoded dataset as ``bytes`` (if successful) or ``None`` if the
        encoding failed.
    """
    # pylint: disable=broad-except
    fp = DicomBytesIO()
    fp.is_implicit_VR = is_implicit_vr
    fp.is_little_endian = is_little_endian
    try:
        write_dataset(fp, ds)
    except Exception as ex:
        LOGGER.error("pydicom.write_dataset() failed:")
        LOGGER.error(ex)
        fp.close()
        return None

    bytestring = fp.parent.getvalue()
    fp.close()

    return bytestring

def encode_element(elem, is_implicit_vr=True, is_little_endian=True):
    """Encode a pydicom DataElement `elem` to bytes.

    Parameters
    ----------
    elem : pydicom.dataelem.DataElement
        The element to encode.
    is_implicit_vr : bool, optional
        The element encoding scheme the element will be encoded with, default
        True.
    is_little_endian : bool, optional
        The byte ordering the element will be encoded in, default True.

    Returns
    -------
    bytes
        The encoded element.
    """
    fp = DicomBytesIO()
    fp.is_implicit_VR = is_implicit_vr
    fp.is_little_endian = is_little_endian
    write_data_element(fp, elem)
    bytestring = fp.parent.getvalue()
    fp.close()

    return bytestring



"""Dummy DIMSE-C SCPs for use in unit tests"""

import logging
import os
import socket
import time
import threading

from pydicom import read_file
from pydicom.dataset import Dataset
from pydicom.uid import UID, ImplicitVRLittleEndian

from pynetdicom import AE, VerificationPresentationContexts
from pynetdicom.sop_class import (
    DisplaySystemSOPClass,
    VerificationSOPClass,
    PrintJobSOPClass,
)
from pynetdicom.status import code_to_category
from .dummy_c_scp import DummyBaseSCP


LOGGER = logging.getLogger('pynetdicom')
LOGGER.setLevel(logging.CRITICAL)
#LOGGER.setLevel(logging.DEBUG)


class DummyEventReportSCP(DummyBaseSCP):
    """A threaded dummy event report SCP used for testing"""
    def __init__(self, port=11112):
        self.ae = AE()
        self.port = port
        self.ae.add_supported_context(PrintJobSOPClass)
        self.ae.add_supported_context(VerificationSOPClass)
        DummyBaseSCP.__init__(self)
        self.status = 0x0000
        ds = Dataset()
        ds.PatientName = 'Test'
        ds.SOPClassUID = PrintJobSOPClass
        ds.SOPInstanceUID = '1.2.3.4'
        self.dataset = ds

    def on_n_event_report(self, event_info, context, info):
        """Callback for ae.on_n_event_report"""

        self.context = context
        self.info = info
        self.event_info = event_info
        time.sleep(self.delay)

        return self.status, self.dataset


class DummyGetSCP(DummyBaseSCP):
    """A threaded dummy get SCP used for testing"""
    def __init__(self, port=11112):
        self.ae = AE()
        self.port = port
        self.ae.add_supported_context(DisplaySystemSOPClass)
        self.ae.add_supported_context(VerificationSOPClass)
        DummyBaseSCP.__init__(self)
        self.status = 0x0000
        ds = Dataset()
        ds.PatientName = 'Test'
        ds.SOPClassUID = DisplaySystemSOPClass
        ds.SOPInstanceUID = '1.2.3.4'
        self.dataset = ds

    def on_n_get(self, attr, context, info):
        """Callback for ae.on_n_get"""

        self.context = context
        self.info = info
        self.attr = attr
        time.sleep(self.delay)

        return self.status, self.dataset


class DummySetSCP(DummyBaseSCP):
    """A threaded dummy get SCP used for testing"""
    def __init__(self, port=11112):
        self.ae = AE()
        self.port = port
        self.ae.add_supported_context(PrintJobSOPClass)
        self.ae.add_supported_context(VerificationSOPClass)
        DummyBaseSCP.__init__(self)
        self.status = 0x0000
        ds = Dataset()
        ds.PatientName = 'Test'
        ds.SOPClassUID = PrintJobSOPClass
        ds.SOPInstanceUID = '1.2.3.4'
        self.dataset = ds

    def on_n_set(self, ds, context, info):
        """Callback for ae.on_n_get"""

        self.context = context
        self.info = info
        time.sleep(self.delay)

        return self.status, self.dataset


class DummyActionSCP(DummyBaseSCP):
    """A threaded dummy get SCP used for testing"""
    def __init__(self, port=11112):
        self.ae = AE()
        self.port = port
        self.ae.add_supported_context(PrintJobSOPClass)
        self.ae.add_supported_context(VerificationSOPClass)
        DummyBaseSCP.__init__(self)
        self.status = 0x0000
        ds = Dataset()
        ds.PatientName = 'Test'
        ds.SOPClassUID = PrintJobSOPClass
        ds.SOPInstanceUID = '1.2.3.4'
        self.dataset = ds

    def on_n_action(self, ds, context, info):
        """Callback for ae.on_n_action"""

        self.context = context
        self.info = info
        time.sleep(self.delay)

        return self.status, self.dataset


class DummyCreateSCP(DummyBaseSCP):
    """A threaded dummy get SCP used for testing"""
    def __init__(self, port=11112):
        self.ae = AE()
        self.port = port
        self.ae.add_supported_context(PrintJobSOPClass)
        self.ae.add_supported_context(VerificationSOPClass)
        DummyBaseSCP.__init__(self)
        self.status = 0x0000
        ds = Dataset()
        ds.PatientName = 'Test'
        ds.SOPClassUID = PrintJobSOPClass
        ds.SOPInstanceUID = '1.2.3.4'
        self.dataset = ds

    def on_n_create(self, ds, context, info):
        """Callback for ae.on_n_create"""
        self.context = context
        self.info = info
        self.ds = ds
        time.sleep(self.delay)

        return self.status, self.dataset


class DummyDeleteSCP(DummyBaseSCP):
    """A threaded dummy delete SCP used for testing"""
    def __init__(self, port=11112):
        self.ae = AE()
        self.port = port
        # Print Job SOP Class
        self.ae.add_supported_context(PrintJobSOPClass)
        self.ae.add_supported_context(VerificationSOPClass)
        DummyBaseSCP.__init__(self)
        self.status = 0x0000

    def on_n_delete(self, context, info):
        """Callback for ae.on_n_delete"""
        self.context = context
        self.info = info
        time.sleep(self.delay)

        return self.status



"""Performance tests for sending DIMSE-C messages."""

import os
import threading
import time

from pydicom import dcmread

from pynetdicom import AE
from pynetdicom.sop_class import CTImageStorage, VerificationSOPClass
from pynetdicom.tests.dummy_c_scp import (
    DummyVerificationSCP, DummyStorageSCP, DummyFindSCP, DummyBaseSCP,
    DummyGetSCP, DummyMoveSCP
)


DS_DIR = os.path.join(os.path.dirname(__file__), '../tests', 'dicom_files')
DATASET = dcmread(os.path.join(DS_DIR, 'CTImageStorage.dcm'))


class TestSendCEcho(object):
    def setup(self):
        """Run prior to each test"""
        self.scp = DummyVerificationSCP()
        self.scp.start()

        ae = AE()
        ae.add_requested_context(VerificationSOPClass)
        self.assoc = ae.associate('localhost', 11112)

    def teardown(self):
        """Clear any active threads"""
        if self.scp:
            self.scp.abort()

        time.sleep(0.1)

        for thread in threading.enumerate():
            if isinstance(thread, DummyBaseSCP):
                thread.abort()
                thread.stop()

    def time_send_c_echo(self):
        "Test sending 100 C-ECHO messages over the same association."
        if self.assoc.is_established:
            for ii in range(100):
                rsp = self.assoc.send_c_echo()
                assert rsp.Status == 0x0000

            self.assoc.release()
        else:
            raise RuntimeError('Unable to associate with the echo SCP')


class TestSendCStore(object):
    def setup(self):
        """Run prior to each test"""
        self.scp = DummyStorageSCP()
        self.scp.status = 0x0000
        self.scp.start()

        ae = AE()
        ae.add_requested_context(CTImageStorage)
        self.assoc = ae.associate('localhost', 11112)

    def teardown(self):
        """Clear any active threads"""
        if self.scp:
            self.scp.abort()

        time.sleep(0.1)

        for thread in threading.enumerate():
            if isinstance(thread, DummyBaseSCP):
                thread.abort()
                thread.stop()

    def time_send_c_store(self):
        "Test sending 100 C-STORE messages over the same association."
        if self.assoc.is_established:
            for ii in range(100):
                rsp = self.assoc.send_c_store(DATASET)
                assert rsp.Status == 0x0000

            self.assoc.release()
        else:
            raise RuntimeError('Unable to associate with the echo SCP')



"""Implementation of the Presentation service."""
from collections import namedtuple
import logging

from pydicom.uid import UID

from pynetdicom.sop_class import (
    _VERIFICATION_CLASSES,
    _STORAGE_CLASSES,
    _QR_CLASSES,
    _BASIC_WORKLIST_CLASSES,
    _RELEVANT_PATIENT_QUERY_CLASSES,
    _SUBSTANCE_ADMINISTRATION_CLASSES,
    _NON_PATIENT_OBJECT_CLASSES,
    _HANGING_PROTOCOL_CLASSES,
    _DEFINED_PROCEDURE_CLASSES,
    _COLOR_PALETTE_CLASSES,
    _IMPLANT_TEMPLATE_CLASSES,
    #_PRINT_MANAGEMENT_CLASSES,
    #_PROCEDURE_STEP_CLASSES,
    _DISPLAY_SYSTEM_CLASSES,
    #_MEDIA_STORAGE_CLASSES,
    #_UNIFIED_PROCEDURE_STEP_CLASSES,
    #_RT_MACHINE_VERIFICATION_CLASSES,
)
from pynetdicom._globals import DEFAULT_TRANSFER_SYNTAXES


LOGGER = logging.getLogger('pynetdicom.presentation')


# Used with the on_c_* callbacks to give the users access to the context
PresentationContextTuple = namedtuple('PresentationContextTuple',
                                      ['context_id',
                                       'abstract_syntax',
                                       'transfer_syntax'])


DEFAULT_ROLE = (True, False, False, True)
BOTH_SCU_SCP_ROLE = (True, True, True, True)
CONTEXT_REJECTED = (False, False, False, False)
INVERTED_ROLE = (False, True, True, False)
SCP_SCU_ROLES = {
    # Req role, ac role, Outcome
    # No SCP/SCU Role Selection proposed
    (None, None) : {
        (None, None) : DEFAULT_ROLE,
        (None, True) : DEFAULT_ROLE,
        (None, False) : DEFAULT_ROLE,
        (True, None) : DEFAULT_ROLE,
        (False, None) : DEFAULT_ROLE,
        (True, True) : DEFAULT_ROLE,
        (True, False) : DEFAULT_ROLE,
        (False, False) : DEFAULT_ROLE,
        (False, True) : DEFAULT_ROLE,
    },
    (True, True) : {
        (None, None) : DEFAULT_ROLE,
        (None, True) : DEFAULT_ROLE,
        (None, False) : DEFAULT_ROLE,
        (True, None) : DEFAULT_ROLE,
        (False, None) : DEFAULT_ROLE,
        (True, True) : BOTH_SCU_SCP_ROLE,
        (True, False) : DEFAULT_ROLE,
        (False, False) : CONTEXT_REJECTED,
        (False, True) : INVERTED_ROLE,
    },
    (True, False) : {
        (None, None) : DEFAULT_ROLE,
        (None, True) : DEFAULT_ROLE,
        (None, False) : DEFAULT_ROLE,
        (True, None) : DEFAULT_ROLE,
        (False, None) : DEFAULT_ROLE,
        (True, True) : DEFAULT_ROLE,  # Invalid
        (True, False) : DEFAULT_ROLE,
        (False, False) : CONTEXT_REJECTED,
        (False, True) : CONTEXT_REJECTED,  # Invalid
    },
    (False, True) : {
        (None, None) : DEFAULT_ROLE,
        (None, True) : DEFAULT_ROLE,
        (None, False) : DEFAULT_ROLE,
        (True, None) : DEFAULT_ROLE,
        (False, None) : DEFAULT_ROLE,
        (True, True) : INVERTED_ROLE,  # Invalid
        (True, False) : CONTEXT_REJECTED,  # Invalid
        (False, False) : CONTEXT_REJECTED,
        (False, True) : INVERTED_ROLE,
    },
    # False, False proposed x
    (False, False) : {
        (None, None) : DEFAULT_ROLE,
        (None, True) : DEFAULT_ROLE,
        (None, False) : DEFAULT_ROLE,
        (True, None) : DEFAULT_ROLE,
        (False, None) : DEFAULT_ROLE,
        (True, True) : CONTEXT_REJECTED,  # Invalid
        (True, False) : CONTEXT_REJECTED,  # Invalid
        (False, False) : CONTEXT_REJECTED,
        (False, True) : CONTEXT_REJECTED,  # Invalid
    },
}


class PresentationContext(object):
    """A Presentation Context primitive.

    **Rules**

    - Each Presentation Context (request) contains:

      - One context ID, an odd integer between 1 and 255.
      - One abstract syntax.
      - One or more transfer syntaxes.
    - Each Presentation Context (response) contains:

      - One context ID, corresponding to a Presentation Context received from
        the Requestor
      - A result, one of 0x00 (acceptance), 0x01 (user rejection), 0x02
        (provider rejection), 0x03 (abstract syntax not supported) or 0x04
        (transfer syntaxes not supported).
      - If the result is 0x00, then a transfer syntax.
      - If any other result, then a transfer syntax may or may not be present.
    - If the result is not 0x00 then the transfer syntax in the reply is not
      significant.
    - The same abstract syntax can be present in more than one Presententation
      Context.
    - Only one transfer syntax can be accepted per Presentation Context.
    - The Presentation Contexts may be sent by the Requestor in any order.
    - The Presentation Contexts may be sent by the Acceptor in any order.

    **SCP/SCU Role Selection Negotiation**

    - If no role selection negotiation then the requestor is SCU and the
      acceptor is SCP
    - The association acceptor cannot accept a role that has not been
      proposed (i.e. cannot return 1 when the proposed value is 0).
    - The association requestor may be SCP only, SCU only or both SCU and SCP.

    +---------------------+---------------------+-------------------+----------+
    | Requestor           | Acceptor            | Outcome           | Notes    |
    +----------+----------+----------+----------+---------+---------+          |
    | scu_role | scp_role | scu_role | scp_role | Req.    | Acc.    |          |
    +==========+==========+==========+==========+=========+=========+==========+
    | N/A      | N/A      | N/A      | N/A      | SCU     | SCP     | Default  |
    +----------+----------+----------+----------+---------+---------+----------+
    | True     | True     | False    | False    | N/A     | N/A     | Rejected |
    |          |          |          +----------+---------+---------+----------+
    |          |          |          | True     | SCP     | SCU     |          |
    |          |          +----------+----------+---------+---------+----------+
    |          |          | True     | False    | SCU     | SCP     | Default  |
    |          |          |          +----------+---------+---------+----------+
    |          |          |          | True     | SCU/SCP | SCU/SCP |          |
    +----------+----------+----------+----------+---------+---------+----------+
    | True     | False    | False    | False    | N/A     | N/A     | Rejected |
    |          |          +----------+          +---------+---------+----------+
    |          |          | True     |          | SCU     | SCP     | Default  |
    +----------+----------+----------+----------+---------+---------+----------+
    | False    | True     | False    | False    | N/A     | N/A     | Rejected |
    |          |          |          +----------+---------+---------+----------+
    |          |          |          | True     | SCP     | SCU     |          |
    +----------+----------+----------+----------+---------+---------+----------+
    | False    | False    | False    | False    | N/A     | N/A     | Rejected |
    +----------+----------+----------+----------+---------+---------+----------+

    As can be seen from the above table there are four possible outcomes:

    * Requestor is SCU, acceptor is SCP (default roles)
    * Requestor is SCP, acceptor is SCU
    * Requestor and acceptor are both SCU/SCP
    * Requestor and acceptor are neither (context rejected)

    Attributes
    ---------
    abstract_syntax : pydicom.uid.UID or None
        The Presentation Context's *Abstract Syntax*.
    as_scp : bool or None
        If True then the association acceptor can act as SCP for the current
        context, otherwise it cannot. A non-None value is only available
        after association negotiation has been completed.
    as_scu : bool or None
        If True then the association acceptor can act as SCU for the current
        context, otherwise it cannot. A non-None value is only available
        after association negotiation has been completed.
    context_id : int or None
        The Presentation Context's *Context ID*.
    result : int or None
        If part of an A-ASSOCIATE (request) then None. If part of an
        A-ASSOCIATE (response) then one of 0x00, 0x01, 0x02, 0x03, 0x04.
    scp_role : bool or None
        Only used when acting as an association acceptor. If True
        then accept when the SCP role is proposed by the requestor, otherwise
        reject the proposal. If None (default) then no SCP/SCU Role Selection
        reply will be sent and the default roles will be used.
    scu_role : bool
        Only used when acting as an association acceptor. If True
        then accept when the SCU role is proposed by the requestor, otherwise
        reject the proposal. If None (default) then no SCP/SCU Role Selection
        reply will be sent and the default roles will be used.
    transfer_syntax : list of pydicom.uid.UID
        The Presentation Context's *Transfer Syntax(es)*.

    References
    ----------

    * DICOM Standard, Part 7, Annexes
      `D.3.2 <http://dicom.nema.org/medical/dicom/current/output/html/part07.html#sect_D.3.2>`_
      and `D.3.3.4 <http://dicom.nema.org/medical/dicom/current/output/html/part07.html#sect_D.3.3.4>`_
    * DICOM Standard, Part 8, Sections
      `9.3.2.2 <http://dicom.nema.org/medical/dicom/current/output/html/part08.html#sect_9.3.2.2>`_,
      `9.3.3.2 <http://dicom.nema.org/medical/dicom/current/output/html/part08.html#sect_9.3.3.2>`_
      and `Annex B <http://dicom.nema.org/medical/dicom/current/output/html/part08.html#chapter_B>`_
    """
    def __init__(self):
        """Create a new PresentationContext."""
        self._context_id = None
        self._abstract_syntax = None
        self._transfer_syntax = []
        self.result = None

        # Used with SCP/SCU Role Selection negotiation
        self._scu_role = None
        self._scp_role = None

        # Used to track the allowed use of the context
        self._as_scp = None
        self._as_scu = None

    @property
    def abstract_syntax(self):
        """Return the presentation context's *Abstract Syntax* as a UID.

        Returns
        -------
        pydicom.uid.UID
        """
        return self._abstract_syntax

    @abstract_syntax.setter
    def abstract_syntax(self, uid):
        """Set the presentation context's *Abstract Syntax*.

        Parameters
        ----------
        uid : str or bytes or pydicom.uid.UID
            The abstract syntax UIDs
        """
        if isinstance(uid, bytes):
            uid = UID(uid.decode('ascii'))
        elif isinstance(uid, str):
            uid = UID(uid)
        else:
            raise TypeError("'abstract_syntax' must be str or bytes or UID")

        if not uid.is_valid:
            LOGGER.warning("'abstract_syntax' set to a non-conformant UID")

        self._abstract_syntax = uid

    def add_transfer_syntax(self, syntax):
        """Append a transfer syntax to the presentation context.

        Parameters
        ----------
        syntax : pydicom.uid.UID, bytes or str
            The transfer syntax to add to the presentation context.
        """
        if isinstance(syntax, str):
            syntax = UID(syntax)
        elif isinstance(syntax, bytes):
            syntax = UID(syntax.decode('ascii'))
        else:
            LOGGER.error("Attempted to add an invalid transfer syntax")
            return

        # If the transfer syntax is rejected we may add an empty str
        if syntax not in self._transfer_syntax and syntax != '':
            if not syntax.is_valid:
                LOGGER.warning("A non-conformant UID has been added "
                               "to 'transfer_syntax'")
            if not syntax.is_private and not syntax.is_transfer_syntax:
                LOGGER.warning("A UID has been added to 'transfer_syntax' "
                               "that is not a transfer syntax")

            self._transfer_syntax.append(syntax)

    @property
    def as_scp(self):
        """Return True if can act as an SCP for the context."""
        return self._as_scp

    @property
    def as_scu(self):
        """Return True if can act as an SCU for the context."""
        return self._as_scu

    @property
    def as_tuple(self):
        """Return a namedtuple representation of the presentation context.

        Intended to be used when the result is 0x00 (accepted) as only the
        first transfer syntax item is returned in the tuple.

        Returns
        -------
        PresentationContextTuple
            A representation of an accepted presentation context.
        """
        return PresentationContextTuple(
            self.context_id, self.abstract_syntax, self.transfer_syntax[0]
        )

    @property
    def context_id(self):
        """Return the presentation context's *ID* parameter as an int."""
        return self._context_id

    @context_id.setter
    def context_id(self, value):
        """Set the presentation context's *ID* parameter.

        Parameters
        ----------
        value : int
            An odd integer between 1 and 255 (inclusive).
        """
        if value is not None and (not 1 <= value <= 255 or value % 2 == 0):
            raise ValueError("'context_id' must be an odd integer between 1 "
                             "and 255, inclusive")

        self._context_id = value

    def __eq__(self, other):
        """Return True if `self` is equal to `other`."""
        if self is other:
            return True

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__

        return NotImplemented

    # Python 2: Classes defining __eq__ should flag themselves as unhashable
    __hash__ = None

    def __ne__(self, other):
        """Return True if `self` does not equal `other`."""
        return not self == other

    @property
    def scp_role(self):
        """Return True if a proposed SCP role will be accepted."""
        return self._scp_role

    @scp_role.setter
    def scp_role(self, val):
        """Set whether to accept the proposed SCP role (as acceptor).

        Parameters
        ----------
        val : bool or None
            If True then accept if the association requestor proposes
            the SCP role for itself, False to reject the proposal. If None
            (default) then no SCP/SCU Role Selection reply will be sent. If
            either of `scu_role` or `scp_role` is None then both will assumed
            to be.
        """
        if not isinstance(val, (bool, type(None))):
            raise TypeError("`scp_role` must be a bool or None")

        self._scp_role = val

    @property
    def scu_role(self):
        """Return True if a proposed SCU role will be accepted."""
        return self._scu_role

    @scu_role.setter
    def scu_role(self, val):
        """Set whether to accept the proposed SCU role (as acceptor).

        Parameters
        ----------
        val : bool or None
            If True then accept if the association requestor proposes
            the SCU role for itself, False to reject the proposal. If None
            (default) then no SCP/SCU Role Selection reply will be sent. If
            either of `scu_role` or `scp_role` is None then both will assumed
            to be.
        """
        if not isinstance(val, (bool, type(None))):
            raise TypeError("`scu_role` must be a bool or None")

        self._scu_role = val

    @property
    def status(self):
        """Return a descriptive str of the presentation context's *Result*.

        Returns
        -------
        str
            The string representation of the result.
        """
        if self.result is None:
            status = 'Pending'
        elif self.result == 0x00:
            status = 'Accepted'
        elif self.result == 0x01:
            status = 'User Rejected'
        elif self.result == 0x02:
            status = 'Provider Rejected'
        elif self.result == 0x03:
            status = 'Abstract Syntax Not Supported'
        elif self.result == 0x04:
            status = 'Transfer Syntax(es) Not Supported'
        else:
            status = 'Unknown'

        return status

    def __str__(self):
        """String representation of the Presentation Context."""
        s = ''
        if self.context_id is not None:
            s += 'ID: {0!s}\n'.format(self.context_id)

        if self.abstract_syntax is not None:
            s += 'Abstract Syntax: {0!s}\n'.format(self.abstract_syntax.name)

        s += 'Transfer Syntax(es):\n'
        for syntax in self.transfer_syntax[:-1]:
            s += '    ={0!s}\n'.format(syntax.name)
        if self.transfer_syntax:
            s += '    ={0!s}'.format(self.transfer_syntax[-1].name)
        else:
            s += '    (none)'

        if self.result is not None:
            s += '\nResult: {0!s}'.format(self.status)

        if None not in (self.as_scu, self.as_scp):
            if self.as_scu and not self.as_scp:
                s += '\nRole: SCU only'
            elif self.as_scu and self.as_scp:
                s += '\nRole: SCU and SCP'
            elif not self.as_scu and self.as_scp:
                s += '\nRole: SCP only'
            else:
                s += '\nRole: (none)'

        return s

    @property
    def transfer_syntax(self):
        """Return the presentation context's *Transfer Syntaxes* as a list.

        Returns
        -------
        list of pydicom.uid.UID
            The transfer syntaxes.
        """
        return self._transfer_syntax

    @transfer_syntax.setter
    def transfer_syntax(self, syntaxes):
        """Set the presentation context's *Transfer Syntaxes*.

        Parameters
        ----------
        syntaxes : list of (str or bytes or pydicom.uid.UID)
            The transfer syntax UIDs to add to the Presentation Context.
        """
        if not isinstance(syntaxes, list):
            raise TypeError("'transfer_syntax' must be a list")

        self._transfer_syntax = []

        for syntax in syntaxes:
            self.add_transfer_syntax(syntax)


def negotiate_as_acceptor(rq_contexts, ac_contexts, roles=None):
    """Process the Presentation Contexts as an Association acceptor.

    Parameters
    ----------
    rq_contexts : list of PresentationContext
        The Presentation Contexts proposed by the peer. Each item has
        values for Context ID, Abstract Syntax and Transfer Syntax.
    ac_contexts : list of PresentationContext
        The Presentation Contexts supported by the local AE when acting
        as an Association acceptor. Each item has values for Context ID
        Abstract Syntax and Transfer Syntax.
    roles : dict or None
        If the requestor has included one or more SCP/SCU Role Selection
        Negotiation items then this will be a dict of
        {SOP Class UID : (SCU role, SCP role)}, otherwise None (default)

    Returns
    -------
    list of PresentationContext
        The accepted presentation context items, each with a Result value
        a Context ID, an Abstract Syntax and one Transfer Syntax item.
        Items are sorted in increasing Context ID value.
    list of SCP_SCU_RoleSelectionNegotiation
        If `roles` is not None then this is a list of SCP/SCU Role Selection
        Negotiation items that can be sent back to the requestor.
    """
    from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation

    roles = roles or {}
    result_contexts = []
    reply_roles = {}

    # No requestor presentation contexts
    if not rq_contexts:
        return result_contexts, []

    # Acceptor doesn't support any presentation contexts
    if not ac_contexts:
        for rq_context in rq_contexts:
            context = PresentationContext()
            context.context_id = rq_context.context_id
            context.abstract_syntax = rq_context.abstract_syntax
            context.transfer_syntax = [rq_context.transfer_syntax[0]]
            context.result = 0x03
            result_contexts.append(context)
        return result_contexts, []

    # Optimisation notes (for iterating through contexts only, not
    #   including actual context negotiation)
    # - Create dict, use set intersection/difference of dict keys: ~600 us
    # - Create dict, iterate over dict keys: ~400 us
    # - Iterate over lists: ~52000 us

    # Requestor may use the same Abstract Syntax in multiple Presentation
    #   Contexts so we need a more specific key than UID
    requestor_contexts = {
        (cx.context_id, cx.abstract_syntax):cx for cx in rq_contexts
    }
    # Acceptor supported SOP Classes must be unique so we can use UID as
    #   the key
    acceptor_contexts = {cx.abstract_syntax:cx for cx in ac_contexts}

    for (cntx_id, ab_syntax) in requestor_contexts:
        # Convenience variable
        rq_context = requestor_contexts[(cntx_id, ab_syntax)]

        # Create a new PresentationContext item that will store the
        #   results of the negotiation
        context = PresentationContext()
        context.context_id = cntx_id
        context.abstract_syntax = ab_syntax
        context._as_scu = False
        context._as_scp = False

        # Check if the acceptor supports the Abstract Syntax
        if ab_syntax in acceptor_contexts:
            # Convenience variables
            ac_context = acceptor_contexts[ab_syntax]
            ac_roles = (ac_context.scu_role, ac_context.scp_role)
            try:
                rq_roles = roles[ab_syntax]
                has_role = True
            except KeyError:
                rq_roles = (None, None)
                has_role = False

            # Abstract syntax supported so check Transfer Syntax
            for tr_syntax in rq_context.transfer_syntax:
                # If transfer syntax supported then (provisionally) accept
                if tr_syntax in ac_context.transfer_syntax:
                    context.transfer_syntax = [tr_syntax]
                    context.result = 0x00
                    result_contexts.append(context)
                    break

            ## SCP/SCU Role Selection Negotiation
            #   Only for (provisionally) accepted contexts
            if context.result == 0x00:
                if None in ac_roles:
                    # Default roles
                    context._as_scu = False
                    context._as_scp = True
                    # If either aq.scu_role or ac.scp_role is None then
                    #   don't send an SCP/SCU negotiation reply
                    has_role = False
                else:
                    # Use a LUT to make changes to outcomes easier
                    #   also its much simpler than coding if/then branches
                    outcome = SCP_SCU_ROLES[rq_roles][ac_roles]
                    context._as_scu = outcome[2]
                    context._as_scp = outcome[3]

                # If can't act as either SCU nor SCP then reject the context
                if context.as_scu is False and context.as_scp is False:
                    # User rejection
                    context.result = 0x01

            # Need to check against None as 0x00 is a possible value
            if context.result is None:
                # Reject context - transfer syntax not supported
                context.result = 0x04
                context.transfer_syntax = [rq_context.transfer_syntax[0]]
                result_contexts.append(context)
            elif context.result == 0x00 and has_role:
                # Create new SCP/SCU Role Selection Negotiation item
                role = SCP_SCU_RoleSelectionNegotiation()
                role.sop_class_uid = context.abstract_syntax

                # Can't return 0x01 if proposed 0x00
                if rq_roles[0] is False:
                    role.scu_role = False
                else:
                    role.scu_role = ac_context.scu_role

                if rq_roles[1] is False:
                    role.scp_role = False
                else:
                    role.scp_role = ac_context.scp_role

                reply_roles[context.abstract_syntax] = role
        else:
            # Reject context - abstract syntax not supported
            context.result = 0x03
            context.transfer_syntax = [rq_context.transfer_syntax[0]]
            result_contexts.append(context)

    # Sort by presentation context ID
    #   This isn't required by the DICOM Standard but its a nice thing to do
    result_contexts = sorted(result_contexts, key=lambda x: x.context_id)

    # Sort role selection by abstract syntax, also not required but nice
    reply_roles = sorted(reply_roles.values(), key=lambda x: x.sop_class_uid)

    return result_contexts, reply_roles


def negotiate_as_requestor(rq_contexts, ac_contexts, roles=None):
    """Process the Presentation Contexts as an Association requestor.

    The acceptor has processed the requestor's presentation context
    definition list and returned the results. We want to do two things:

    - Process the SCP/SCU Role Selection Negotiation
    - Return a nice list of PresentationContexts with the Results and
      original Abstract Syntax values to make things easier to use.

    Presentation Context Item (RQ)

    - Presentation context ID
    - Abstract Syntax: one
    - Transfer syntax: one or more

    Presentation Context Item (AC)

    - Presentation context ID
    - Result: 0x00, 0x01, 0x02, 0x03, 0x04
    - Transfer syntax: one, not to be tested if result is not 0x00

    Parameters
    ----------
    rq_contexts : list of PresentationContext
        The Presentation Contexts sent to the peer as the A-ASSOCIATE's
        Presentation Context Definition List.
    ac_contexts : list of PresentationContext
        The Presentation Contexts return by the peer as the A-ASSOCIATE's
        Presentation Context Definition Result List.
    roles : dict or None
        If the acceptor has included one or more SCP/SCU Role Selection
        Negotiation items then this will be a dict of
        {SOP Class UID : (SCU role, SCP role)}, otherwise None (default)

    Returns
    -------
    list of PresentationContext
        The contexts in the returned Presentation Context Definition Result
        List, with added Abstract Syntax value. Items are sorted in
        increasing Context ID value and the SCP/SCU roles are set as per
        the negotiated outcome.
    """
    roles = roles or {}

    if not rq_contexts:
        raise ValueError('Requestor contexts are required')
    output = []

    # Create dicts, indexed by the presentation context ID
    requestor_contexts = {
        context.context_id:context for context in rq_contexts
    }
    acceptor_contexts = {
        context.context_id:context for context in ac_contexts
    }

    for context_id in requestor_contexts:
        # Convenience variable
        rq_context = requestor_contexts[context_id]

        context = PresentationContext()
        context.context_id = context_id
        context.abstract_syntax = rq_context.abstract_syntax
        # Ensure we always have a role set
        context._as_scu = False
        context._as_scp = False

        if context_id in acceptor_contexts:
            # Convenience variable
            ac_context = acceptor_contexts[context_id]

            # Update with accepted values
            if ac_context.transfer_syntax:
                context.transfer_syntax = [ac_context.transfer_syntax[0]]
            context.result = ac_context.result

            ## SCP/SCU Role Selection Negotiation
            rq_roles = (rq_context.scu_role, rq_context.scp_role)
            try:
                ac_roles = roles[context.abstract_syntax]
            except KeyError:
                ac_roles = (None, None)

            # Skip if context rejected or acceptor ignored proposal
            if ac_context.result == 0x00 and None not in ac_roles:
                outcome = SCP_SCU_ROLES[rq_roles][ac_roles]
                context._as_scu = outcome[0]
                context._as_scp = outcome[1]
            else:
                # We are the association requestor, so SCU role only
                context._as_scp = False
                context._as_scu = True

        # Add any missing contexts as rejected
        else:
            context.transfer_syntax = [rq_context.transfer_syntax[0]]
            context.result = 0x02

        output.append(context)

    # Sort returned list by context ID
    return sorted(output, key=lambda x: x.context_id)


def build_context(abstract_syntax, transfer_syntax=None):
    """Return a PresentationContext built from the `abstract_syntax`.

    Parameters
    ----------
    abstract_syntax : str or UID or sop_class.SOPClass
        The UID or SOPClass instance to use as the abstract syntax.
    transfer_syntax : str/UID or list of str/UID
        The transfer syntax UID(s) to use (default: [Implicit VR Little Endian,
        Explicit VR Little Endian, Implicit VR Big Endian])

    Examples
    --------

    Specifying a presentation context with the default transfer syntaxes

    >>> from pynetdicom import build_context
    >>> context = build_context('1.2.840.10008.1.1')
    >>> print(context)
    Abstract Syntax: Verification SOP Class
    Transfer Syntax(es):
        =Implicit VR Little Endian
        =Explicit VR Little Endian
        =Explicit VR Big Endian

    Specifying the abstract syntax using a pynetdicom SOPClass instance and
    a single transfer syntax

    >>> from pynetdicom import build_context
    >>> from pynetdicom.sop_class import VerificationSOPClass
    >>> context = build_context(VerificationSOPClass, '1.2.840.10008.1.2')
    >>> print(context)
    Abstract Syntax: Verification SOP Class
    Transfer Syntax(es):
        =Implicit VR Little Endian

    Specifying multiple transfer syntaxes

    >>> from pydicom.uid import UID
    >>> from pynetdicom import build_context
    >>> context = build_context(UID('1.2.840.10008.1.1'),
                                ['1.2.840.10008.1.2',
    ...                          '1.2.840.10008.1.2.4.50'])
    >>> print(context)
    Abstract Syntax: Verification SOP Class
    Transfer Syntax(es):
        =Implicit VR Little Endian
        =JPEG Baseline (Process 1)


    Returns
    -------
    presentation.PresentationContext
    """
    if transfer_syntax is None:
        transfer_syntax = DEFAULT_TRANSFER_SYNTAXES

    abstract_syntax = UID(abstract_syntax)

    # Allow single transfer syntax values for convenience
    if isinstance(transfer_syntax, str):
        transfer_syntax = [transfer_syntax]

    context = PresentationContext()
    context.abstract_syntax = abstract_syntax
    context.transfer_syntax = transfer_syntax

    return context


def build_role(uid, scu_role=False, scp_role=False):
    """Return a SCP/SCU Role Selection Negotiation item.

    Parameters
    ----------
    uid : str or UID or sop_class.SOPClass
        The UID or SOPClass instance to use as the *SOP Class UID* parameter
        value.
    scu_role : bool, optional
        True to propose the SCU role for the *Requestor*, False otherwise
        (default).
    scp_role : bool, optional
        True to propose the SCP role for the *Requestor*, False otherwise
        (default).

    Returns
    -------
    pdu_primitives.SCP_SCU_RoleSelectionNegotiation
        The role selection item.
    """
    from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation

    role = SCP_SCU_RoleSelectionNegotiation()
    role.sop_class_uid = uid
    role.scu_role = scu_role
    role.scp_role = scp_role
    return role


# Service specific pre-generated Presentation Contexts
# pylint: disable=line-too-long,invalid-name
VerificationPresentationContexts = [
    build_context(uid) for uid in sorted(_VERIFICATION_CLASSES.values())
]

StoragePresentationContexts = [
    build_context(uid) for uid in sorted(_STORAGE_CLASSES.values())[:128]
]

QueryRetrievePresentationContexts = [
    build_context(uid) for uid in sorted(_QR_CLASSES.values())
]

BasicWorklistManagementPresentationContexts = [
    build_context(uid) for uid in sorted(_BASIC_WORKLIST_CLASSES.values())
]

RelevantPatientInformationPresentationContexts = [
    build_context(uid) for uid in sorted(_RELEVANT_PATIENT_QUERY_CLASSES.values())
]

SubstanceAdministrationPresentationContexts = [
    build_context(uid) for uid in sorted(_SUBSTANCE_ADMINISTRATION_CLASSES.values())
]

NonPatientObjectPresentationContexts = [
    build_context(uid) for uid in sorted(_NON_PATIENT_OBJECT_CLASSES.values())
]

HangingProtocolPresentationContexts = [
    build_context(uid) for uid in sorted(_HANGING_PROTOCOL_CLASSES.values())
]

DefinedProcedureProtocolPresentationContexts = [
    build_context(uid) for uid in sorted(_DEFINED_PROCEDURE_CLASSES.values())
]

ColorPalettePresentationContexts = [
    build_context(uid) for uid in sorted(_COLOR_PALETTE_CLASSES.values())
]

ImplantTemplatePresentationContexts = [
    build_context(uid) for uid in sorted(_IMPLANT_TEMPLATE_CLASSES.values())
]

DisplaySystemPresentationContexts = [
    build_context(uid) for uid in sorted(_DISPLAY_SYSTEM_CLASSES.values())
]
